/*
 * main.c
 *
 *  Created on: 17.03.2011
 *      Author: AndreR
 *
 * Pin Configuration
 *
 * PB1: CHG / OC1A
 * PB2: CS
 * PB3: MOSI
 * PB4: MISO
 * PB5: SCK
 *
 * PC1: T IGBT 1 / ADC1
 * PC2: CHG-ADC / ADC2
 * PC3: CAP-ADC / ADC3
 * PC4: T IGBT 2 / ADC4
 *
 * PD0: LEDR
 * PD1: LEDG
 * PD3: FAN / OC2B
 * PD5: DIS / OC0B
 * PD6: EN-DIS1
 * PD7: EN-DIS2
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "kicker_data.h"
#include "spi.h"
#include "hal.h"

#define STATE_IDLE		0
#define STATE_FIRE		1
#define STATE_CHARGE	2

#define CHARGE_MODE_MANUAL		0
#define CHARGE_MODE_AUTOLOAD	1

#define NUM_INTERVALLS 6

typedef struct PowerIntervall
{
	uint16_t capVoltIntervall;

	uint16_t on;
	uint16_t off;
} PowerIntervall;

PowerIntervall powerIntervalls[NUM_INTERVALLS];

KickerData data;

volatile uint16_t sysTick;
volatile uint16_t chargeTimer;
volatile uint16_t chargeTimerEnd;
volatile uint8_t fireDone;

uint8_t state;
uint8_t chargeMode;
uint8_t fanMode;

uint16_t maxCapLevel;


ISR(TIMER2_OVF_vect)	// TODO: sideeffects of NOBLOCK?
{
	// 100us, systick
	sysTick++;

	if(chargeTimer != 0xFFFF)
	{
		chargeTimer++;
		if(chargeTimer == chargeTimerEnd)
			DisableCharge();
	}
}

ISR(TIMER1_COMPB_vect)
{
	DisableFire();

	fireDone = 1;
}

int main()
{
	// init ports
	PORTD = (1 << LEDG) | (1 << LEDR);
	DDRD = (1 << LEDG) | (1 << LEDR) | (1 << FAN) | (1 << EN_DIS1) | (1 << EN_DIS2) | (1 << DIS);
	DDRB = (1 << CHG);

	PORTD &= ~(1 << LEDR);

	// init variables
	uint8_t i;

	sysTick = 0;

	state = STATE_IDLE;
	chargeMode = CHARGE_MODE_MANUAL;
	fanMode = KICKER_FAN_MANUAL;
	fireDone = 1;

	uint8_t overChargeCounter = 0;

	chargeTimer = 0xFFFF;

	maxCapLevel = 0;

	uint8_t hitMaxCap = 0;

	// init autoload settings
	powerIntervalls[0].on	= 4;
	powerIntervalls[0].off	= 78;
	powerIntervalls[0].capVoltIntervall=0;
	powerIntervalls[1].on	= 4;
	powerIntervalls[1].off	= 78;
	powerIntervalls[1].capVoltIntervall=300;
	powerIntervalls[2].on	= 10;	// K:  6, S: 12
	powerIntervalls[2].off	= 36;	// K: 36, S: 66
	powerIntervalls[2].capVoltIntervall=1000;
	powerIntervalls[3].on	= 12;	// K:  6, S: 12
	powerIntervalls[3].off	= 30;	// K: 12, S: 30
	powerIntervalls[3].capVoltIntervall=2000;
	powerIntervalls[4].on	= 12;	// K: 6, S: 12
	powerIntervalls[4].off	= 18;	// K: 6, S: 18
	powerIntervalls[4].capVoltIntervall=3000;
	powerIntervalls[5].on	= 12;	// K: 6, S: 12
	powerIntervalls[5].off	= 12;	// K: 6, S: 12
	powerIntervalls[5].capVoltIntervall=4000;

	uint16_t maxChargeCurrent = 5000;	// 4A

	// init SPI
	SPIInit();

	// init ADC
	ADCSRA = (1 << ADEN) | (1 << ADIF) | (1 << ADPS2) | (1 << ADPS1);	// Enable, prescaler 64 = 187.5kHz
	DIDR0 = 0x3F;

	// init Timer 2 (systick + FAN)
	TCCR2A = (1 << COM2B1);	// non-inverting Fast PWM on OC2B (FAN)
	TCCR2A |= (1 << WGM20) | (1 << WGM21);
	TCCR2B = (1 << WGM22);	// Fast PWM mode, OCRA is TOP, TOVF set on TOP
	TCNT2 = 0;	// clear counter
	OCR2A = 150;	// with prescaler 8 makes a 10kHz PWM
	OCR2B = 0;	// initial FAN level is 0
	TIMSK2 = (1 << TOIE2);	// enable overflow interrupt (systick 100us)
	TCCR2B |= (1 << CS21);	// enable

	// init Timer 0 (DIS1 + DIS2)
	TCCR0A = (1 << COM0B1) | (1 << WGM00); // non-inverting Phase Correct PWM on OC0B, TOP = OCR0A
	TCCR0B = (1 << WGM02);
	TCNT0 = 0;
	OCR0A = 240;
	OCR0B = 0;
	TCCR0B |= (1 << CS00);	//enable, prescaler = 1, 25kHz

	// init Timer 1 (CHG)
	TCCR1B = (1 << WGM13);
	TCNT1 = 0;
	OCR1A = 0;
	ICR1 = 12;	// determines period (T), in clock cycles (@12Mhz)
	TCCR1B |= (1 << CS10);	// enable
	TCCR1A = (1 << COM1A1); // non-inverting Phase and frequency correct PWM, TOP = ICR1

	// init data
	data.cmd = 0;
	data.uCap = 0;
	data.iChg = 0;
	data.TIGBT[0] = 0;
	data.TIGBT[1] = 0;

	uint16_t ch[4];

	PORTD |= (1 << LEDR);

	sei();

	while(1)
	{
		// --- Update shared data ---
		SPIExchange(data.exchange.miso, data.exchange.mosi, KICKER_DATA_SIZE);

		// --- Command processing ---
		if(data.cmd == KICKER_CMD_SET_FAN)
		{
			fanMode = data.cmds.setFan.mode;
			if(fanMode == KICKER_FAN_MANUAL)
			{
				SetFanSpeed(data.cmds.setFan.speed);
			}
		}

		if(data.cmd == KICKER_CMD_FIRE && state != STATE_FIRE && data.cmds.fire.duration.value != 0)
		{
			state = STATE_FIRE;
			DisableCharge();

			PORTD &= ~(1 << LEDR);

			fireDone = 0;

			if(data.cmds.fire.device == KICKER_DEVICE_STRAIGHT)
			{
				EnableFire1(data.cmds.fire.tOn, data.cmds.fire.duration);
			}

			if(data.cmds.fire.device == KICKER_DEVICE_CHIP)
			{
				EnableFire2(data.cmds.fire.tOn, data.cmds.fire.duration);
			}
		}

		if(data.cmd == KICKER_CMD_LOAD_AUTO)
		{
			maxCapLevel = data.cmds.loadAuto.maxCapLevel;
			if(maxCapLevel > 3600)	// 360V
				maxCapLevel = 3600;

			if(maxCapLevel < 50)
			{
				maxCapLevel = 50;
				chargeMode = CHARGE_MODE_MANUAL;
			}
			else
			{
				chargeMode = CHARGE_MODE_AUTOLOAD;
			}
		}

		if(data.cmd == KICKER_CMD_LOAD_MANUAL && state != STATE_FIRE)
		{
			chargeMode = CHARGE_MODE_MANUAL;

			chargeTimerEnd = data.cmds.loadManual.duration;
			chargeTimer = 0;

			EnableCharge(data.cmds.loadManual.tOnTicks, data.cmds.loadManual.tOffTicks);
		}

		data.cmd = 0;

		// --- FAN speed handling ---
		// TODO: implement automatic mode

		// --- State Machine ---
		if(state == STATE_FIRE)
		{
			if(fireDone)
			{
				state = STATE_IDLE;
				PORTD |= (1 << LEDR);
			}
		}

		if(data.uCap > 3600)
		{
			DisableCharge();
			chargeMode = CHARGE_MODE_MANUAL;
		}

		if(state == STATE_IDLE && chargeMode == CHARGE_MODE_AUTOLOAD)
		{

			// automatic loading
			if(hitMaxCap == 0)
			{
				if(data.uCap <= maxCapLevel)
				{
					for(i = 0; i< NUM_INTERVALLS; i++)
					{
						if(data.uCap <= powerIntervalls[i].capVoltIntervall)
							break;
					}

					EnableCharge(powerIntervalls[i].on, powerIntervalls[i].off);
				}
				else
				{
					hitMaxCap = 1;
					DisableCharge();
				}
			}
			else
			{
				// hit max cap, wait 5V
				if(data.uCap <= (maxCapLevel - 50))
				{
					hitMaxCap = 0;
				}
			}
		}

		// charge current limiter
		if(data.iChg > maxChargeCurrent)
		{
			overChargeCounter++;
			DisableFire();
			DisableCharge();
			fireDone = 1;

			if(overChargeCounter > 100)
			{
				overChargeCounter = 0;
				DisableCharge();
				chargeMode = CHARGE_MODE_MANUAL;
			}
		}
		else
		{
			overChargeCounter = 0;
		}

		// --- ADC Sampling ---
		for(i = 0; i < 4; i++)
			ch[i] = 0;

		uint8_t iterations;
		for(iterations = 0; iterations < 10; iterations++)	// ~5ms overall
		{
			for(i = 0; i < 4; i++)
			{
				ADMUX = i+1;
				ADCSRA |= (1 << ADSC);
				while(ADCSRA & (1 << ADSC));	// ~78us
				ch[i] += ADC;
			}
		}

		for(i = 0; i < 4; i++)
		{
			ch[i] /= iterations;
			ch[i] *= 20;	// convert to [Ve-4]
		}

		// Temperature conversion from LM60 sensors
		uint16_t temp[2];
		temp[0] = ch[0];
		temp[1] = ch[3];

		uint32_t val;
		for(i = 0; i < 2; i++)
		{
			val = temp[i];
			val -= 4240;	// -424mV (offset)
			val *= 100;		// [Ve-4] => [Ve-6]
			val /= 625;	// [Ve-6] => [Ce-1]
			temp[i] = val;
		}

		data.TIGBT[0] = temp[0];
		data.TIGBT[1] = temp[1];

		// capacitor voltage conversion
		val = ch[2];	// CAP
		val *= 196;	// [Ve-4] => [Ve-7], Resistor divider gain = 196
		val /= 1000;	// [Ve-7] => [Ve-4]
		data.uCap = val;

		// charge current conversion
		val = ch[1];	// CHG
		val *= 100;	// [Ve-4] => [Ve-6]
		val /= 40;	// Gain 40
		val /= 5;	// [Ve-6]/[Ohm*e-3] => [A*e-3]
		data.iChg = val;

		// --- Sys LED (green) ---
		if(sysTick >= 10000)
		{
			sysTick = 0;
			PORTD ^= (1 << LEDG);
		}
	}

	return 0;
}

