#include "stm32f10x.h"
#include "misc.h"
#include "stm32f10x_rcc.h"
#include "stm32f10x_exti.h"
#include "stm32f10x_gpio.h"
#include "stm32f10x_flash.h"
#include "stm32f10x_usart.h"
#include "stm32f10x_iwdg.h"

#include "core_cm3.h"

#include "util/bit_banding.h"

#include "FreeRTOS.h"
#include "task.h"

#include "usb/usb.h"
#include "network/dect_interface.h"
#include "cli.h"
#include "network.h"
#include "ir.h"
#include "spi1_2.h"
#include "movement.h"
#include "power.h"
#include "hal/led.h"
#include "acceleration.h"
#include "kicker.h"
#include "util/console.h"

#include "hal/motor.h"
#include "hal/enc.h"
#include "hal/buzzer.h"
#include "hal/exti.h"
#include "hal/usart1.h"

#include "init.h"

void RCC_Configuration();
void NVIC_Configuration();

void SystemInit()
{
}

int main()
{
	RCC_Configuration();

	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_DMA1, ENABLE);
	RCC_APB2PeriphClockCmd(	RCC_APB2Periph_GPIOA | RCC_APB2Periph_GPIOB | RCC_APB2Periph_GPIOC |
							RCC_APB2Periph_GPIOD | RCC_APB2Periph_GPIOE | RCC_APB2Periph_GPIOF |
							RCC_APB2Periph_GPIOG | RCC_APB2Periph_AFIO, ENABLE);

//	NVIC->ISER;
	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA | SCB_SHCSR_BUSFAULTENA | SCB_SHCSR_MEMFAULTENA;

	NVIC_Configuration();

	PowerPreInit();
	PowerEnable5VSupply(1);

	// Configure and start Independent watchdog
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
	IWDG_SetPrescaler(IWDG_Prescaler_8);
	IWDG_SetReload(2500);	// approx. 500ms
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);

	// Initialize shared resources
	// --- HAL ---
	EncInit();
	MotorInit();
	BuzzerInit();
	ExtiInit();
	USBInit();
	DECTInterfaceInit();
	USART1Init();

	// --- GATEKEEPERS ---
	USART1Init();
	SPI1_2Init();

	// --- High-Level Systems ---
	ConsoleInit();
	LEDInit();
	PowerInit();
	MovementInit();
	AccelerationInit();
	IRInit();
	NetworkInit();
	KickerInit();

	LEDSet(0xFF);

	xTaskCreate(InitTask, (const signed char* const)"Init", 256, NULL, 3, NULL);

	vTaskStartScheduler();

	return 0;
}

void RCC_Configuration()
{
	RCC_DeInit();

	RCC_HSEConfig(RCC_HSE_ON);
	while(RCC_GetFlagStatus(RCC_FLAG_HSERDY) == RESET);

	FLASH_PrefetchBufferCmd(FLASH_PrefetchBuffer_Enable);
	FLASH_SetLatency(FLASH_Latency_2);

	RCC_PLLConfig(RCC_PLLSource_HSE_Div1, RCC_PLLMul_6);
	RCC_PLLCmd(ENABLE);

	RCC_HCLKConfig(RCC_SYSCLK_Div1);	// set AHB speed
	// setting APB2 to 72Mhz leads to some SPI problems (missing last bit)
	RCC_PCLK2Config(RCC_HCLK_Div2);		// set high speed APB2 (36Mhz)
	RCC_PCLK1Config(RCC_HCLK_Div2);		// set low speed APB1 (36Mhz)

//	RCC_PCLK2Config(RCC_HCLK_Div8);		// set high speed APB2 (9Mhz)
//	RCC_PCLK1Config(RCC_HCLK_Div8);		// set low speed APB1 (9Mhz)

	RCC_SYSCLKConfig(RCC_SYSCLKSource_PLLCLK);
	while(RCC_GetSYSCLKSource() != 0x08);

	SysTick_CLKSourceConfig(SysTick_CLKSource_HCLK);
}

void NVIC_Configuration()
{
	NVIC_SetVectorTable(NVIC_VectTab_FLASH, 0x00);

	NVIC_PriorityGroupConfig(NVIC_PriorityGroup_4);
}

