/*
 * dect.c
 *
 *  Created on: 01.07.2010
 *      Author: AndreR
 */

#include "network.h"

#include "stm32f10x.h"
#include "misc.h"
#include "stm32f10x_rcc.h"
#include "stm32f10x_gpio.h"
#include "stm32f10x_usart.h"

#include "task.h"

#include "util/bit_banding.h"

#include "util/console.h"
#include "errors.h"
#include "util/mem.h"
#include "commands.h"
#include "movement.h"
#include "kicker.h"
#include "ir.h"
#include "acceleration.h"
#include "network/dect_interface.h"
#include "network/ethernet.h"
#include "network/arp.h"
#include "network/ipv4.h"
#include "network/udp.h"
#include "network/igmpv3.h"

Network network;

void DECTRxCallback();
uint8_t processPacket(PacketHeader* pHeader, uint8_t* pBuf, FrameHeader* pFrameHeader, IPv4Header* pIPHeader, UDPHeader* pUDPHeader);

int16_t NetworkSendARP(uint8_t operation, MAC targetMAC, IPv4Address targetIP);

void verboseOutput();
void printMAC(MAC mac);
void printIP(IPv4Address ip);

void NetworkInit()
{
	network.state = NETWORK_STATE_HEADLESS;
	network.ethernetIdent = 0;

	network.my.mac = MACSet(2, 0, 0, 0, 0, 0);
	network.my.ip = IPv4AddressSet(192, 168, 255, 254);	// IP must be valid, dst discards packet otherwise
	network.my.port = 0;

	network.ucast.mac = MACSet(2, 0, 0, 0, 0, 0);
	network.ucast.ip = IPv4AddressSet(0, 0, 0, 0);
	network.ucast.port = 0;

	network.mcast.mac = MACSet(0x01, 0x00, 0x5E, 0x2A, 0x2A, 0x2A);
	network.mcast.ip = IPv4AddressSet(225, 42, 42, 42);
	network.mcast.port = 10041;	// listen on this port
	network.mcast.serverPort = 10040;	// send to this one

	vSemaphoreCreateBinary(network.hNewPacket);
	vSemaphoreCreateBinary(network.hSendPacket);

	DECTInterfaceSetRxCallback(&DECTRxCallback);

	network.cpuId[0] = *((uint32_t*)(0x1FFFF7E8 + 0));
	network.cpuId[1] = *((uint32_t*)(0x1FFFF7E8 + 4));
	network.cpuId[2] = *((uint32_t*)(0x1FFFF7E8 + 8));

	network.botId = 250;

	network.igmp.respCounter = 0;
	network.igmp.respTime = 500;
	network.igmp.lastSendTime = 0;
	network.igmp.robustness = 2;

	network.enableTimeout = 1;
}

void DECTRxCallback()
{
	signed portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;

	xSemaphoreGiveFromISR(network.hNewPacket, &xHigherPriorityTaskWoken);

	if(xHigherPriorityTaskWoken == pdTRUE)
		taskYIELD();
}

void stopAll()
{
	MovementSetAutomatic(0, 0, 0, 0);

	for(uint8_t i = 0; i < 4; i++)
	{
		MovementSetPidSetpoint(i, 0);
		MovementSetManual(i, 0);
	}

	MovementSetDribble(0);

	IRDisarm();

	KickerLoadManual(6, 78, 1);
	vTaskDelay(200/portTICK_RATE_MS);
	KickerFire(10000 | (3 << 14), 240, KICKER_DEVICE_STRAIGHT);

	network.state = NETWORK_STATE_HEADLESS;

	network.my.mac = MACSet(2, 0, 0, 0, 0, 0);
	network.my.ip = IPv4AddressSet(192, 168, 255, 254);
	network.my.port = 0;

	network.ucast.mac = MACSet(2, 0, 0, 0, 0, 0);
	network.ucast.ip = IPv4AddressSet(0, 0, 0, 0);
	network.ucast.port = 0;

	network.botId = 250;
}

void NetworkTask(void* params)
{
	params = params;

	portTickType xLastPacketTime = xTaskGetTickCount();
	portTickType xLastStatsTime = xLastPacketTime;

	int16_t result = 0;

	uint8_t* buf = network.rxProcessingBuf;

	{
		PacketHeader header;
		header.cmd = CMD_SYSTEM_ANNOUNCEMENT;
		header.len = LEN_SYSTEM_ANNOUNCEMENT;
		header.section = SECTION_SYSTEM;

		SystemAnnouncement announce;
		announce.cpuId[0] = network.cpuId[0];
		announce.cpuId[1] = network.cpuId[1];
		announce.cpuId[2] = network.cpuId[2];

		result = NetworkSendPacket(&header, (uint8_t*)&announce, 1);

		ConsolePrint("Send initial announcement %d\r\n", result);
	}

	while(1)
	{
		if(DECTInterfaceGetMode() == DECT_MODE_DATA)
		{
			xSemaphoreTake(network.hNewPacket, 10/portTICK_RATE_MS);

			uint8_t* pDecoded;
			uint16_t decodedLength;

			// update stats
			if(xTaskGetTickCount() - xLastStatsTime > 1000/portTICK_RATE_MS)
			{
				xLastStatsTime = xTaskGetTickCount();

				network.lastStats.rx.bytes = dectInterface.stats.rx.bytes;
				network.lastStats.rx.packets = dectInterface.stats.rx.packets;
				network.lastStats.rx.drops = dectInterface.stats.rx.drops;
				network.lastStats.tx.bytes = dectInterface.stats.tx.bytes;
				network.lastStats.tx.packets = dectInterface.stats.tx.packets;
				network.lastStats.tx.drops = dectInterface.stats.tx.drops;

				if(network.lastStats.rx.drops > 0)
					ConsolePrint("Dropped %d last second\r\n", network.lastStats.rx.drops);

				DECTResetStats();
			}

			if(xTaskGetTickCount() - xLastPacketTime > 1000/portTICK_RATE_MS && network.enableTimeout
					&& network.state == NETWORK_STATE_ASSIGNED)
			{
				// TIMEOUT!

				stopAll();
			}

			// IGMPv3 work
			if(network.igmp.respTime != 0)
			{
				if(xTaskGetTickCount() - network.igmp.lastSendTime > network.igmp.respTime/portTICK_RATE_MS)
				{
					network.igmp.lastSendTime = xTaskGetTickCount();

					IGMPv3GroupRecord record;
					record.auxDataLen = 0;
					record.numberOfSources = 0;
					record.recordType = IGMPV3_MODE_IS_EXCLUDE;
					record.multicastAddress.u32 = network.mcast.ip.u32;

					NetworkSendIGMPv3MembershipReport(1, record._data, IGMPV3_GROUP_RECORD_SIZE);

					network.igmp.respCounter++;
					if(network.igmp.respCounter >= network.igmp.robustness)
						network.igmp.respTime = 0;
				}
			}

			// process network stack
			result = DECTInterfaceGet(buf, NETWORK_RX_PROCESSING_SIZE, &pDecoded, &decodedLength);
			if(result != ERROR_FIFO_PACKET_SYNCED_EMPTY && result != 0)
				ConsolePrint("DECTInterface RX error\r\n");
			if(result)
				continue;


//			ConsolePrint("\r\nRecv'd HDLC data, len: %d\r\n", decodedLength);

			FrameHeader* pFrameHeader;

			result = EthernetDecode(pDecoded, decodedLength, &pDecoded, &pFrameHeader);
			if(result)
			{
				continue;
			}

			if(pFrameHeader->dstMAC.cast.multicast == 0 && MACCompare(pFrameHeader->dstMAC, network.my.mac) == 0)
				continue; // we are not addressed

//			ConsolePrint("A packet for us :)\r\n");

			if(pFrameHeader->type == ETHERNET_PROTOCOL_ARP)
			{
				ARPPacketIPv4* pARP;

				if(ARPDecode(pDecoded, ARP_PACKET_SIZE, &pARP))
					continue;

//				ConsolePrint("ARP\r\n");

				if(	pARP->operation == ARP_OPERATION_REQUEST && pARP->dstProtAddr.u32 == network.my.ip.u32 &&
					network.state == NETWORK_STATE_ASSIGNED)
				{
					// someone wants to know our MAC address
//					ConsolePrint("Replying to ARP!\r\n");

					NetworkSendARP(ARP_OPERATION_REPLY, pARP->srcHwAddr, pARP->srcProtAddr);
					continue;
				}
			}

			if(pFrameHeader->type == ETHERNET_PROTOCOL_IPV4)
			{
				IPv4Header* pIPHeader;

				if(IPv4Decode(pDecoded, decodedLength, &pDecoded, &pIPHeader))
					continue;

				if(	pIPHeader->dstIP.u32 != network.mcast.ip.u32 && pIPHeader->dstIP.u32 != network.my.ip.u32 &&
					pIPHeader->dstIP.u32 != IPv4AddressSet(224, 0, 0, 1).u32)
					continue;	// we are not addressed

				if(pIPHeader->protocol == IPV4_PROTOCOL_UDP)
				{
					UDPHeader* pUDPHeader;

					if(UDPDecode(pDecoded, pIPHeader->totalLength - pIPHeader->ihl*4, &pDecoded, &pUDPHeader))
						continue;

					if(pUDPHeader->dstPort != network.mcast.port && pUDPHeader->dstPort != network.my.port)
						continue;

//					ConsolePrint("Valid UDP packet for us\r\n");

					if(processPacket((PacketHeader*)pDecoded, pDecoded + PACKET_HEADER_SIZE, pFrameHeader, pIPHeader, pUDPHeader))
						xLastPacketTime = xTaskGetTickCount();
				}

				if(pIPHeader->protocol == IPV4_PROTOCOL_IGMP)
				{
					uint8_t type = 0;
					IGMPv3GetType(pDecoded, &type);

					if(type == IGMP_MEMBERSHIP_QUERY)
					{
						IGMPv3MembershipQuery* pQuery;
						IPv4Address* pSourceList;

						result = IGMPv3DecodeMembershipQuery(pDecoded, pIPHeader->totalLength - pIPHeader->ihl*4, &pQuery, &pSourceList);
						if(result)
							continue;

						if(pQuery->groupAddress.u32 == 0 || pQuery->groupAddress.u32 == network.mcast.ip.u32)
						{
							// general query
							network.igmp.lastSendTime = xTaskGetTickCount();
							network.igmp.respCounter = 0;
							network.igmp.robustness = pQuery->QRV ? pQuery->QRV : 2;
							uint32_t respTime = IGMPv3DecodeMaxRespCode(pQuery->maxRespCode);
							network.igmp.respTime = xTaskGetTickCount() % respTime;
						}
					}

					if(type == IGMPV3_MEMBERSHIP_REPORT)
					{

					}
				}

				if(pIPHeader->protocol == IPV4_PROTOCOL_ICMP)
				{

				}
			}
		}
		else
		{
			vTaskDelay(10/portTICK_RATE_MS);
		}
	}
}

int16_t NetworkSendPacket(PacketHeader* pHeader, uint8_t* pData, uint8_t multicast)
{
	if(network.state == NETWORK_STATE_HEADLESS && multicast == 0)
		return ERROR_NETWORK_HEADLESS;

	if(pHeader->len + PACKET_HEADER_SIZE > NETWORK_TX_PROCESSING_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	xSemaphoreTake(network.hSendPacket, portMAX_DELAY);	// prevent re-entering

	MemCpy(pHeader->_data, network.txProcessingBuf, PACKET_HEADER_SIZE);
	MemCpy(pData, network.txProcessingBuf + PACKET_HEADER_SIZE, pHeader->len);

	uint16_t encLength = pHeader->len + PACKET_HEADER_SIZE;

	uint16_t dstPort = network.ucast.port;
	if(multicast)
		dstPort = network.mcast.serverPort;

	IPv4Address dstIp = network.ucast.ip;
	if(multicast)
		dstIp = network.mcast.ip;

	MAC dstMAC = network.ucast.mac;
	if(multicast)
		dstMAC = network.mcast.mac;

	int16_t result = UDPEncode(network.txProcessingBuf, NETWORK_TX_PROCESSING_SIZE, encLength, &encLength, network.my.port, dstPort);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = IPv4Encode(network.txProcessingBuf, NETWORK_TX_PROCESSING_SIZE, encLength, &encLength,
						network.ethernetIdent, network.my.ip, dstIp, IPV4_PROTOCOL_UDP);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = EthernetEncode(network.txProcessingBuf, NETWORK_TX_PROCESSING_SIZE, encLength, &encLength,
							network.my.mac, dstMAC, ETHERNET_PROTOCOL_IPV4);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = DECTInterfacePut(network.txProcessingBuf, encLength);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	network.ethernetIdent++;

	xSemaphoreGive(network.hSendPacket);

	return 0;
}

int16_t NetworkSendIGMPv3MembershipReport(uint16_t numberOfGroupRecords, uint8_t* recordData, uint16_t recordDataLength)
{
	if(recordDataLength + IGMPV3_MEMBERSHIP_REPORT_SIZE > NETWORK_TX_PROCESSING_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	xSemaphoreTake(network.hSendPacket, portMAX_DELAY);	// prevent re-entering

	uint16_t encLength = 0;
	int16_t result = 0;

	result = IGMPv3EncodeMembershipReport(	network.txProcessingBuf, NETWORK_TX_PROCESSING_SIZE,
											&encLength, numberOfGroupRecords, recordData, recordDataLength);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	IPv4Address srcIP = IPv4AddressSet(0, 0, 0, 0);
	IPv4Address dstIP = IPv4AddressSet(224, 0, 0, 22);
	MAC dstMAC = MACSet(0x01, 0x00, 0x5E, 0x00, 0x00, 0x16);	// 224.0.0.22

	result = IPv4EncodeWithRouterAlertOption(network.txProcessingBuf, NETWORK_TX_PROCESSING_SIZE, encLength, &encLength,
						network.ethernetIdent, srcIP, dstIP, IPV4_PROTOCOL_IGMP);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = EthernetEncode(network.txProcessingBuf, NETWORK_TX_PROCESSING_SIZE, encLength, &encLength,
							network.my.mac, dstMAC, ETHERNET_PROTOCOL_IPV4);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = DECTInterfacePut(network.txProcessingBuf, encLength);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	network.ethernetIdent++;

	xSemaphoreGive(network.hSendPacket);

	return 0;
}

/*
int16_t NetworkSendICMP(uint8_t* pData, uint16_t dataLength, IPv4Address dstIP, MAC dstMAC)
{
	if(dataLength > NETWORK_TX_PROCESSING_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	if(network.state == NETWORK_STATE_HEADLESS)
		return ERROR_NETWORK_HEADLESS;

	xSemaphoreTake(network.hSendPacket, portMAX_DELAY);	// prevent re-entering

	int16_t result = 0;

	result = IPv4Encode(network.txProcessingBuf, NETWORK_TX_PROCESSING_SIZE, dataLength, &dataLength,
						network.ethernetIdent, network.my.ip, dstIP, IPV4_PROTOCOL_ICMP);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = EthernetEncode(network.txProcessingBuf, NETWORK_TX_PROCESSING_SIZE, dataLength, &dataLength,
							network.my.mac, dstMAC, ETHERNET_PROTOCOL_IPV4);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = DECTInterfacePut(network.txProcessingBuf, dataLength);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	network.ethernetIdent++;

	xSemaphoreGive(network.hSendPacket);

	return 0;
}
*/

int16_t NetworkSendARP(uint8_t operation, MAC targetMAC, IPv4Address targetIP)
{
	if(network.state == NETWORK_STATE_HEADLESS)
		return ERROR_NETWORK_HEADLESS;

	if(ARP_PACKET_SIZE > NETWORK_TX_PROCESSING_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	xSemaphoreTake(network.hSendPacket, portMAX_DELAY);	// prevent re-entering

	uint16_t encLength;

	int16_t result = ARPEncode(	network.txProcessingBuf, NETWORK_TX_PROCESSING_SIZE, &encLength,
								operation, network.my.mac, network.my.ip, targetMAC, targetIP);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = EthernetEncode(network.txProcessingBuf, NETWORK_TX_PROCESSING_SIZE, encLength, &encLength,
							network.my.mac, targetMAC, ETHERNET_PROTOCOL_ARP);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = DECTInterfacePut(network.txProcessingBuf, encLength);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	xSemaphoreGive(network.hSendPacket);

	return 0;
}


void verboseOutput()
{
	uint8_t* buf = network.rxProcessingBuf;
	uint8_t* pDecoded;
	uint16_t decodedLength;

	if(!DECTInterfaceGet(buf, NETWORK_RX_PROCESSING_SIZE, &pDecoded, &decodedLength))
	{
		ConsolePrint("\r\nRecv'd HDLC data, len: %d\r\n", decodedLength);

		FrameHeader* pFrameHeader;

		int16_t result = EthernetDecode(pDecoded, decodedLength, &pDecoded, &pFrameHeader);
		if(result)
		{
			ConsolePrint("Ethernet error: 0x%X\r\n", result);
			return;
		}

		ConsolePrint("Dst MAC: ");
		printMAC(pFrameHeader->dstMAC);
		ConsolePrint("\r\nSrc MAC: ");
		printMAC(pFrameHeader->srcMAC);
		ConsolePrint("\r\nType: 0x%X\r\n", pFrameHeader->type, pFrameHeader->type);

		if(pFrameHeader->type == ETHERNET_PROTOCOL_ARP)
		{
			ARPPacketIPv4* pARP;

			if(ARPDecode(pDecoded, ARP_PACKET_SIZE, &pARP))
				return;

			ConsolePrint("--- ARP ---\r\n");
			ConsolePrint("OP: %d\r\n", pARP->operation);
			ConsolePrint("Src MAC: ");
			printMAC(pARP->srcHwAddr);
			ConsolePrint("\r\nSrc IP: ");
			printIP(pARP->srcProtAddr);
			ConsolePrint("\r\nDst MAC: ");
			printMAC(pARP->dstHwAddr);
			ConsolePrint("\r\nDst IP: ");
			printIP(pARP->dstProtAddr);
			ConsolePrint("\r\n");
		}

		if(pFrameHeader->type == ETHERNET_PROTOCOL_IPV4)
		{
			IPv4Header* pIPHeader;

			if(IPv4Decode(pDecoded, decodedLength, &pDecoded, &pIPHeader))
				return;

			ConsolePrint("--- IPv4 ---\r\n");
			ConsolePrint("Length: %d\r\n", pIPHeader->totalLength);
			ConsolePrint("TTL: %d\r\n", pIPHeader->ttl);
			ConsolePrint("Protocol: %d\r\n", pIPHeader->protocol);
			ConsolePrint("Src IP: ");
			printIP(pIPHeader->srcIP);
			ConsolePrint("\r\nDst IP: ");
			printIP(pIPHeader->dstIP);
			ConsolePrint("\r\n");

			if(pIPHeader->protocol == IPV4_PROTOCOL_UDP)
			{
				UDPHeader* pUDPHeader;

				if(UDPDecode(pDecoded, pIPHeader->totalLength - pIPHeader->ihl*4, &pDecoded, &pUDPHeader))
					return;

				ConsolePrint("--- UDP ---\r\n");
				ConsolePrint("Src Port: %d\r\n", pUDPHeader->srcPort);
				ConsolePrint("Dst Port: %d\r\n", pUDPHeader->dstPort);
				ConsolePrint("Length: %d\r\n", pUDPHeader->length);
				ConsolePrint("CRC: 0x%X\r\n", pUDPHeader->crc);
			}
		}
	}
}


void printMAC(MAC mac)
{
	ConsolePrint("%X-%X-%X-%X-%X-%X", mac.u8[0], mac.u8[1], mac.u8[2], mac.u8[3], mac.u8[4], mac.u8[5]);
}

void printIP(IPv4Address ip)
{
	ConsolePrint("%d.%d.%d.%d", ip.u8[0], ip.u8[1], ip.u8[2], ip.u8[3]);
}

uint8_t processPacket(PacketHeader* pHeader, uint8_t* pBuf, FrameHeader* pFrameHeader, IPv4Header* pIPHeader, UDPHeader* pUDPHeader)
{
	pUDPHeader = pUDPHeader;	// avoid warning

	switch(pHeader->section)
	{
		case SECTION_SYSTEM:
		{
			switch(pHeader->cmd)
			{
				case CMD_SYSTEM_SET_IDENTITY:
				{
					ConsolePrint("Received identity command\r\n");

					SystemSetIdentity* pIdent = (SystemSetIdentity*)pBuf;

					if(pIdent->cpuId[0] == network.cpuId[0] && pIdent->cpuId[1] == network.cpuId[1] &&
						pIdent->cpuId[2] == network.cpuId[2])
					{
						// this bot is addressed!

						if(pIdent->botId > 11)	// check if this bot is commanded to reset
						{
							ConsolePrint("Changing to headless mode\r\n");

							network.state = NETWORK_STATE_HEADLESS;

							network.my.mac = MACSet(2, 0, 0, 0, 0, 0);
							network.my.ip = IPv4AddressSet(192, 168, 255, 254);
							network.my.port = 0;

							network.ucast.mac = MACSet(2, 0, 0, 0, 0, 0);
							network.ucast.ip = IPv4AddressSet(0, 0, 0, 0);
							network.ucast.port = 0;

							stopAll();
						}
						else
						{
							ConsolePrint("Changing to assigned mode\r\n");

							MemCpy(pIdent->mac, network.my.mac.u8, 6);
							MemCpy(pIdent->ip, network.my.ip.u8, 4);
							network.my.port = pIdent->port;

							MemCpy(pFrameHeader->srcMAC.u8, network.ucast.mac.u8, 6);
							MemCpy(pIPHeader->srcIP.u8, network.ucast.ip.u8, 4);
							network.ucast.port = pIdent->serverPort;
							network.botId = pIdent->botId;

							network.state = NETWORK_STATE_ASSIGNED;
						}

						return 1;
					}
				}
				break;
				case CMD_SYSTEM_PING:
				{
					SystemPing* pPing = (SystemPing*)pBuf;

					SystemPong pong;
					pong.id = pPing->id;

					PacketHeader header;
					header.section = SECTION_SYSTEM;
					header.cmd = CMD_SYSTEM_PONG;
					header.len = LEN_SYSTEM_PONG;

					NetworkSendPacket(&header, (uint8_t*)&pong, 0);

					return 1;
				}
				break;
				case CMD_SYSTEM_BIG_PING:
				{
					SystemBigPing* pPing = (SystemBigPing*)pBuf;

					SystemPong pong;
					pong.id = pPing->id;

					PacketHeader header;
					header.section = SECTION_SYSTEM;
					header.cmd = CMD_SYSTEM_PONG;
					header.len = LEN_SYSTEM_PONG;

					NetworkSendPacket(&header, (uint8_t*)&pong, 0);

					return 1;
				}
				break;
				case CMD_SYSTEM_SET_LOGS:
				{
					SystemSetLogs* pLogs = (SystemSetLogs*)pBuf;

					for(uint8_t i = 0; i < 5; i++)
					{
						MovementSetLog(i, pLogs->motor[i]);
					}

					MovementSetFeedbackLog(pLogs->movement);
					KickerSetLogging(pLogs->kicker);
					AccelerationSetLogging(pLogs->accel);
					IRSetLog(pLogs->ir);

					return 1;
				}
				break;
			}
		}
		break;
		case SECTION_DECT:
		{
		}
		break;
		case SECTION_KICKER:
		{
			switch(pHeader->cmd)
			{
				case CMD_KICKER_KICKV2:
				{
					KickerKickV2* pKick = (KickerKickV2*)pBuf;

					if(pKick->mode == KICKER_MODE_FORCE)
					{
						KickerFire(pKick->duration, pKick->level, pKick->device);
					}

					if(pKick->mode == KICKER_MODE_ARM)
					{
						IRArm(pKick->device, pKick->duration, pKick->level);
					}

					if(pKick->mode == KICKER_MODE_DISARM)
					{
						IRDisarm();
					}

					return 1;
				}
				break;
				case CMD_KICKER_CHARGE_MANUAL:
				{
					KickerChargeManual* pCharge = (KickerChargeManual*)pBuf;

					KickerLoadManual(pCharge->tOnTicks, pCharge->tOffTicks, pCharge->duration);

					return 1;
				}
				break;
				case CMD_KICKER_CHARGE_AUTO:
				{
					KickerChargeAuto* pCharge = (KickerChargeAuto*)pBuf;

					KickerLoadAuto(pCharge->maxCapLevel);

					return 1;
				}
				break;
			}
		}
		break;
		case SECTION_MOTOR:
		{
			switch(pHeader->cmd)
			{
				case CMD_MOTOR_MOVE:
				{
					MotorMove* pMove = (MotorMove*)pBuf;

					MovementSetAutomatic(pMove->vx, pMove->vy, pMove->vw, 0);

					return 1;
				}
				break;
				case CMD_MOTOR_MOVE_V2:
				{
					MotorMoveV2* pMove = (MotorMoveV2*)pBuf;

					MovementSetAutomatic(pMove->vx, pMove->vy, pMove->vw, pMove->vv);

					return 1;
				}
				break;
				case CMD_MOTOR_DRIBBLE:
				{
					MotorDribble* pDribble = (MotorDribble*)pBuf;

					MovementSetDribble(pDribble->rpm);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_PID_PARAMS:
				{
					MotorSetPidParams* pParams = (MotorSetPidParams*)pBuf;

					MovementSetMotorPidParams(pParams->id, pParams->Kp, pParams->Ki, pParams->Kd, pParams->slewMax);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_MODE:
				{
					MotorSetMode* pMode = (MotorSetMode*)pBuf;

					ConsolePrint("MotorSetMode\r\n");

					MovementSetMode(pMode->mode);
					MovementSetLog(pMode->id, pMode->log);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_MANUAL:
				{
					MotorSetManual* pManual = (MotorSetManual*)pBuf;

					MovementSetManual(pManual->id, pManual->power);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_PID_SP:
				{
					MotorSetPidSp* pSp = (MotorSetPidSp*)pBuf;

					MovementSetPidSetpoint(pSp->id, pSp->setpoint);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_PARAMS:
				{
					MotorSetParams* pParams = (MotorSetParams*)pBuf;

					MovementSetMode(pParams->mode);

					for(uint8_t i = 0; i < 5; i++)
					{
						MovementSetMotorPidParams(i, pParams->Kp[i], pParams->Ki[i], pParams->Kd[i], pParams->slewMax[i]);
					}

					return 1;
				}
				break;
			}
		}
		break;
		case SECTION_MULTICAST:
		{
			switch(pHeader->cmd)
			{
				case CMD_MULTICAST_UPDATE_ALL:
				{
					MulticastUpdateAll* pAll = (MulticastUpdateAll*)pBuf;

					for(uint8_t i = 0; i < 5; i++)
					{
						if(pAll->botId[i] == network.botId)
						{
							MotorMove* pMove = &pAll->move[i];
							KickerKickV2* pKick = &pAll->kick[i];
							MotorDribble* pDribble = &pAll->dribble[i];

							MovementSetAutomatic(pMove->vx, pMove->vy, pMove->vw, 0);

							MovementSetDribble(pDribble->rpm);

							if(pKick->mode == KICKER_MODE_FORCE)
							{
								KickerFire(pKick->duration, pKick->level, pKick->device);
							}

							if(pKick->mode == KICKER_MODE_ARM)
							{
								IRArm(pKick->device, pKick->duration, pKick->level);
							}

							if(pKick->mode == KICKER_MODE_DISARM)
							{
								IRDisarm();
							}

							return 1;
						}
					}
				}
				break;

				case CMD_MULTICAST_UPDATE_ALL_V2:
				{
					MulticastUpdateAllV2* pAll = (MulticastUpdateAllV2*)pBuf;

					for(uint8_t i = 0; i < 5; i++)
					{
						if(pAll->botId[i] == network.botId)
						{
							MotorMoveV2* pMove = &pAll->move[i];
							KickerKickV2* pKick = &pAll->kick[i];
							MotorDribble* pDribble = &pAll->dribble[i];

							if(pAll->useMove & (1 << i))
								MovementSetAutomatic(pMove->vx, pMove->vy, pMove->vw, pMove->vv);

							MovementSetDribble(pDribble->rpm);

							if(pKick->mode == KICKER_MODE_FORCE)
							{
								KickerFire(pKick->duration, pKick->level, pKick->device);
							}

							if(pKick->mode == KICKER_MODE_ARM)
							{
								IRArm(pKick->device, pKick->duration, pKick->level);
							}

							if(pKick->mode == KICKER_MODE_DISARM)
							{
								IRDisarm();
							}

							return 1;
						}
					}
				}
				break;
			}
		}
		break;
		case SECTION_MOVEMENT:
		{
			switch(pHeader->cmd)
			{
				case CMD_MOVEMENT_SET_PARAMS:
				{
				}
				break;
			}
		}
		break;
	}

	return 0;
}
