/*
 * atmega.c
 *
 *  Created on: 09.03.2011
 *      Author: AndreR
 */

#include "atmega.h"
#include "errors.h"
#include "util/mem.h"
#include "util/critical.h"

#include "task.h"

#define CMD_PROGRAM 2
#define CMD_PROGRAM_NO_RX 3
#define CMD_EXCHANGE 4

#define START_FLAG	0x3C
#define END_FLAG	0xA5

int16_t prepareForData(ATMega* pAT, uint16_t timeout)
{
	xSemaphoreGive(pAT->hWriteSemaphore);
	if(xSemaphoreTake(pAT->hWriteSemaphore, timeout/portTICK_RATE_MS) == pdFALSE)
	{
		return ERROR_ATMEGA_TIMEOUT;
	}

	return 0;
}

int16_t waitForData(ATMega* pAT, uint16_t timeout)
{
	if(xSemaphoreTake(pAT->hWriteSemaphore, timeout/portTICK_RATE_MS) == pdFALSE)
		return ERROR_ATMEGA_TIMEOUT;

	return 0;
}

int16_t deviceLookup(uint8_t* pSig, uint16_t* pFlashSize, uint16_t* pFlashPageSize)
{
	if(pSig[0] != 0x1E)
		return ERROR_ATMEGA_UNKOWN_SIGNATURE;

	if(pSig[1] == 0x92 && pSig[2] == 0x0A)	// ATmega48P
	{
		*pFlashSize = 2048;
		*pFlashPageSize = 32;
		return 0;
	}

	if(pSig[1] == 0x93 && pSig[2] == 0x0F)	// ATmega88P
	{
		*pFlashSize = 4096;
		*pFlashPageSize = 32;
		return 0;
	}

	if(pSig[1] == 0x94 && pSig[2] == 0x0B)	// ATmega168P
	{
		*pFlashSize = 8192;
		*pFlashPageSize = 64;
		return 0;
	}

	return ERROR_ATMEGA_UNKOWN_SIGNATURE;
}

void ATMegaCallback(SPIHeader* pHeader, uint8_t* pData, void* pUserData)
{
	ATMega* pAT = (ATMega*)pUserData;

	if(pHeader->slaveId != pAT->slaveId)
		return;

	if(pHeader->cmd == CMD_EXCHANGE)
	{
		MemCpy(pData, pAT->rx, pHeader->length);

		xSemaphoreGive(pAT->hWriteSemaphore);
	}

	if(pHeader->cmd == CMD_PROGRAM)
	{
		MemCpy(pData, pAT->rx, pHeader->length);

		xSemaphoreGive(pAT->hWriteSemaphore);
	}

	if(pHeader->cmd == CMD_PROGRAM_NO_RX)
	{
		xSemaphoreGive(pAT->hWriteSemaphore);
	}
}

int16_t ATMegaInit(ATMega* pAT, SPI* pSPI, GPIO_TypeDef* pCSPort, uint16_t csPin, GPIO_TypeDef* pResetPort, uint16_t resetPin)
{
	pAT->pSPI = pSPI;
	pAT->bufValid = 0;
	pAT->programming = 0;
	pAT->pResetPort = pResetPort;
	pAT->resetPin = resetPin;
	pAT->signature[0] = 0;
	pAT->signature[1] = 0;
	pAT->signature[2] = 0;
	pAT->fuses[0] = 0xFF;
	pAT->fuses[1] = 0xFF;
	pAT->fuses[2] = 0xFF;

	pResetPort->BSRR = resetPin;

	pAT->hDataInUse = xSemaphoreCreateMutex();
	if(pAT->hDataInUse == NULL)
		return ERROR_INVALID_HANDLE;

	vSemaphoreCreateBinary(pAT->hWriteSemaphore);
	if(pAT->hWriteSemaphore == NULL)
		return ERROR_INVALID_HANDLE;

	SPISlave slave;
	slave.callback = &ATMegaCallback;
	slave.pUserData = pAT;
	slave.pCSPort = pCSPort;
	slave.csPin = csPin;

	slave.prescaler = SPI_BaudRatePrescaler_256;
	slave.cpol = 0;	// idle low
	slave.cpha = 0;	// sample on first clock transition (ATMega mode 0)

	return SPIAddSlave(pSPI, slave, &pAT->slaveId);
}

/*
 * Blocking Function!
 */
int16_t ATMegaExchange(ATMega* pAT, uint8_t* pMosiData, uint8_t* pMisoData, uint16_t dataLength)
{
	if(pAT->programming)
		return ERROR_ATMEGA_MODE_FAULT;

	dataLength /= 2;

	int16_t result = 0;

	TASK_MUTEX(pAT->hDataInUse)
	{
		pAT->tx[0] = START_FLAG;
		MemCpy(pMosiData, pAT->tx+1, dataLength);

		// CRC calculation
		uint16_t i;
		uint8_t crc = 0;
		for(i = 0; i < dataLength; i++)
		{
			crc ^= pMosiData[i];
		}

		pAT->tx[dataLength+1] = crc;
		pAT->tx[dataLength+2] = END_FLAG;

		// Build SPI packet
		SPIHeader header;
		header.cmd = CMD_EXCHANGE;
		header.length = dataLength+3;
		header.slaveId = pAT->slaveId;

		// Prepare to send (clear mutex)
		result = prepareForData(pAT, 1000);
		if(result)
			return result;

		// Send packet
		result = SPIPutPacket(pAT->pSPI, &header, dataLength+3, pAT->tx);
		if(result)
			return result;

		// wait until data is updated
		result = waitForData(pAT, 1000);
		if(result)
			return result;

		// FLAG check
		if(pAT->rx[0] != START_FLAG || pAT->rx[dataLength+2] != END_FLAG)
			return ERROR_ATMEGA_EXCHANGE_FLAG_MISMATCH;

		// crc check
		crc = 0;
		for(i = 1; i < dataLength+1; i++)
		{
			crc ^= pAT->rx[i];
		}

		if(crc == pAT->rx[i])
		{
			// CRC good
			MemCpy(pAT->rx+1, pMisoData, dataLength);
		}
		else
			result = ERROR_ATMEGA_EXCHANGE_CRC_WRONG;
	}

	return result;
}

int16_t ATMegaEnterProgramming(ATMega* pAT)
{
	// try to enter exclusive SPI mode
	while(SPISetExclusiveSlave(pAT->pSPI, pAT->slaveId) != 0)
	{
		vTaskDelay(10/portTICK_RATE_MS);
	}

	// wait until spi processing is idle
	while(!SPIIdle(pAT->pSPI))
	{
		vTaskDelay(10/portTICK_RATE_MS);
	}

	int16_t result = SPISetExclusiveConfig(pAT->pSPI);
	if(result)
		return result;

	pAT->programming = 1;
	pAT->bufValid = 0;
	uint8_t counter;

	TASK_MUTEX(pAT->hDataInUse)
	{
		uint8_t* pTx = pAT->tx;

		result = prepareForData(pAT, 1000);
		if(result)
		{
			pAT->programming = 0;
			return result;
		}

		SPIHeader header;
		header.cmd = CMD_PROGRAM;
		header.slaveId = pAT->slaveId;

		counter = 0;

		while(counter < 100)
		{
			pAT->pResetPort->BRR = pAT->resetPin;	// assert reset
			vTaskDelay(20/portTICK_RATE_MS);

			// Build "Enable Serial Instruction" command
			pTx[0] = 0xAC;
			pTx[1] = 0x53;
			pTx[2] = 0x00;
			pTx[3] = 0x00;
			pAT->rx[2] = 0x00;

			header.length = 4;

			SPIPutPacket(pAT->pSPI, &header, 4, pTx);

			waitForData(pAT, 20);

			counter++;

			if(pAT->rx[2] != 0x53)
			{
				pAT->pResetPort->BSRR = pAT->resetPin;
				vTaskDelay(10/portTICK_RATE_MS);
				continue;
			}
			else
			{
				break;
			}
		}
	}

	if(counter >= 100) // not synced
	{
		pAT->pResetPort->BSRR = pAT->resetPin;
		pAT->programming = 0;
		SPIReleaseExclusiveSlave(pAT->pSPI);
		xSemaphoreGive(pAT->hWriteSemaphore);

		return ERROR_ATMEGA_OUT_OF_SYNC;
	}

	return 0;
}

int16_t ATMegaExitProgramming(ATMega* pAT)
{
	if(pAT->programming == 0)
		return 0;

	xSemaphoreGive(pAT->hWriteSemaphore);

	// wait until spi processing is idle
	while(!SPIIdle(pAT->pSPI))
	{
		vTaskDelay(10/portTICK_RATE_MS);
	}

	pAT->pResetPort->BSRR = pAT->resetPin;

	SPIReleaseExclusiveSlave(pAT->pSPI);

	pAT->programming = 0;

	return 0;
}

int16_t ATMegaReadSignature(ATMega* pAT)
{
	if(pAT->programming == 0)
		return ERROR_ATMEGA_MODE_FAULT;

	TASK_MUTEX(pAT->hDataInUse)
	{
		if(prepareForData(pAT, 100))
			return ERROR_ATMEGA_TIMEOUT;

		// Build "Read signature byte" commands
		uint8_t* pTx = pAT->tx;

		pTx[0] = 0x30;
		pTx[1] = 0x00;
		pTx[2] = 0x00;
		pTx[3] = 0x00;

		pTx[4] = 0x30;
		pTx[5] = 0x00;
		pTx[6] = 0x01;
		pTx[7] = 0x00;

		pTx[8] = 0x30;
		pTx[9] = 0x00;
		pTx[10] = 0x02;
		pTx[11] = 0x00;

		SPIHeader header;
		header.cmd = CMD_PROGRAM;
		header.slaveId = pAT->slaveId;
		header.length = 12;

		SPIPutPacket(pAT->pSPI, &header, 12, pTx);

		// wait for data
		if(waitForData(pAT, 100))
			return ERROR_ATMEGA_TIMEOUT;

		pAT->signature[0] = pAT->rx[3];
		pAT->signature[1] = pAT->rx[7];
		pAT->signature[2] = pAT->rx[11];
	}

	return 0;
}

int16_t ATMegaReadFuses(ATMega* pAT)
{
	if(pAT->programming == 0)
		return ERROR_ATMEGA_MODE_FAULT;

	TASK_MUTEX(pAT->hDataInUse)
	{
		if(prepareForData(pAT, 100))
			return ERROR_ATMEGA_TIMEOUT;

		uint8_t* pTx = pAT->tx;

		// read low fuse
		pTx[0] = 0x50;
		pTx[1] = 0x00;
		pTx[2] = 0x00;
		pTx[3] = 0x00;

		// read high fuse
		pTx[4] = 0x58;
		pTx[5] = 0x08;
		pTx[6] = 0x00;
		pTx[7] = 0x00;

		// read extended fuse
		pTx[8] = 0x50;
		pTx[9] = 0x08;
		pTx[10] = 0x00;
		pTx[11] = 0x00;

		SPIHeader header;
		header.cmd = CMD_PROGRAM;
		header.slaveId = pAT->slaveId;
		header.length = 12;

		SPIPutPacket(pAT->pSPI, &header, 12, pTx);

		// wait for data
		if(waitForData(pAT, 100))
			return ERROR_ATMEGA_TIMEOUT;

		pAT->fuses[0] = pAT->rx[3];
		pAT->fuses[1] = pAT->rx[7];
		pAT->fuses[2] = pAT->rx[11];
	}

	return 0;
}

int16_t ATMegaWriteFuses(ATMega* pAT, uint8_t low, uint8_t high, uint8_t extended)
{
	if(pAT->programming == 0)
		return ERROR_ATMEGA_MODE_FAULT;

	TASK_MUTEX(pAT->hDataInUse)
	{
		if(prepareForData(pAT, 100))
			return ERROR_ATMEGA_TIMEOUT;

		uint8_t* pTx = pAT->tx;

		// write low fuse
		pTx[0] = 0xAC;
		pTx[1] = 0xA0;
		pTx[2] = 0x00;
		pTx[3] = low;

		// write high fuse
		pTx[4] = 0xAC;
		pTx[5] = 0xA8;
		pTx[6] = 0x00;
		pTx[7] = high;

		// write extended fuse
		pTx[8] = 0xAC;
		pTx[9] = 0xA4;
		pTx[10] = 0x00;
		pTx[11] = extended;

		SPIHeader header;
		header.cmd = CMD_PROGRAM_NO_RX;
		header.slaveId = pAT->slaveId;
		header.length = 12;

		SPIPutPacket(pAT->pSPI, &header, 12, pTx);

		// wait for data
		if(waitForData(pAT, 100))
			return ERROR_ATMEGA_TIMEOUT;
	}

	return 0;
}

int16_t ATMegaChipErase(ATMega* pAT)
{
	if(pAT->programming == 0)
		return ERROR_ATMEGA_MODE_FAULT;

	TASK_MUTEX(pAT->hDataInUse)
	{
		if(prepareForData(pAT, 100))
			return ERROR_ATMEGA_TIMEOUT;

		// Build "Chip Erase" command
		uint8_t* pTx = pAT->tx;

		pTx[0] = 0xAC;
		pTx[1] = 0x80;
		pTx[2] = 0x00;
		pTx[3] = 0x00;

		SPIHeader header;
		header.cmd = CMD_PROGRAM;
		header.slaveId = pAT->slaveId;
		header.length = 4;

		SPIPutPacket(pAT->pSPI, &header, 4, pTx);

		// wait for data
		if(waitForData(pAT, 100))
			return ERROR_ATMEGA_TIMEOUT;

		vTaskDelay(10/portTICK_RATE_MS);
	}

	return 0;
}

int16_t ATMegaLoadProgram(ATMega* pAT, const uint8_t* pData, uint16_t dataSize)	// dataSize in bytes!
{
	if(pAT->programming == 0)
		return ERROR_ATMEGA_MODE_FAULT;

	if(pAT->signature[0] != 0x1E)
		return ERROR_ATMEGA_UNKOWN_SIGNATURE;

	TASK_MUTEX(pAT->hDataInUse)
	{
		if(prepareForData(pAT, 100))
			return ERROR_ATMEGA_TIMEOUT;

		dataSize /= 2;	// dataSize now in words!
		uint16_t flashSize = 0;	// unit: words!
		uint16_t flashPageSize = 0;	// unit: words!

		int16_t result = deviceLookup(pAT->signature, &flashSize, &flashPageSize);
		if(result)
			return result;

		if(dataSize > flashSize*2)
			return ERROR_ATMEGA_PROGRAM_TOO_BIG;

		SPIHeader header;
		header.cmd = CMD_PROGRAM_NO_RX;
		header.slaveId = pAT->slaveId;

		uint8_t* pBuf = pAT->buf;
		uint16_t bufPos = 0;
		uint16_t readLocation = 0;

		uint16_t numPagesToWrite = dataSize/flashPageSize;
		if(dataSize % flashPageSize != 0)
			numPagesToWrite++;

		for(uint16_t page = 0; page < numPagesToWrite; page++)
		{
			uint16_t writeLocation = page*flashPageSize;
			uint16_t numWordsToWrite = flashPageSize;

			if(page == numPagesToWrite-1)	// last page?
				numWordsToWrite = dataSize - writeLocation;

			bufPos = 0;

			//load page
			for( uint16_t i = 0; i < numWordsToWrite; i++ )
			{
				//1. Byte: Low Byte 0x40
				pBuf[bufPos+0] = 0x40;
				//2. Byte: don't care
				pBuf[bufPos+1] = 0x00;
				//3. Byte: Word Adress
				pBuf[bufPos+2] = i;
				//4. Byte: databyte
				pBuf[bufPos+3] = pData[readLocation];

				//1. Byte: High Byte 0x48
				pBuf[bufPos+4] = 0x48;
				//2. Byte: don't care
				pBuf[bufPos+5] = 0x00;
				//3. Byte: Word Adress
				pBuf[bufPos+6] = i;
				//4. Byte: databyte
				pBuf[bufPos+7] = pData[readLocation+1];

				readLocation += 2;
				bufPos += 8;
			}

			/* Speicherung der Buffer Page*/
			//"Write Program Memory Page"
			pBuf[bufPos+0] = 0x4C;
			//Adress High Bits
			pBuf[bufPos+1] = (writeLocation >> 8);
			//3. Byte: Word Adress
			pBuf[bufPos+2] = (writeLocation & 0xFF);
			//4. Byte: databyte
			pBuf[bufPos+3] = 0;

			bufPos += 4;

			header.length = bufPos;

			result = SPIPutPacket(pAT->pSPI, &header, bufPos, pBuf);
			if(result)
				return result;

			if(waitForData(pAT, 2000))
				return ERROR_ATMEGA_TIMEOUT;

			/*wait min. 4.5 ms*/
			vTaskDelay(10/portTICK_RATE_MS);
		}
	}

	return 0;
}

int16_t ATMegaVerifyProgram(ATMega* pAT, const uint8_t* pData, uint16_t dataSize)	// dataSize in bytes!
{
	if(pAT->programming == 0)
		return ERROR_ATMEGA_MODE_FAULT;

	if(pAT->signature[0] != 0x1E)
		return ERROR_ATMEGA_UNKOWN_SIGNATURE;

	TASK_MUTEX(pAT->hDataInUse)
	{
		if(prepareForData(pAT, 100))
			return ERROR_ATMEGA_TIMEOUT;

		dataSize /= 2;	// dataSize now in words!
		uint16_t flashSize = 0;	// unit: words!
		uint16_t flashPageSize = 0;	// unit: words!

		int16_t result = deviceLookup(pAT->signature, &flashSize, &flashPageSize);
		if(result)
			return result;

		if(dataSize > flashSize*2)
			return ERROR_ATMEGA_PROGRAM_TOO_BIG;

		SPIHeader header;
		header.cmd = CMD_PROGRAM;
		header.slaveId = pAT->slaveId;

		uint8_t* pBuf = pAT->tx;

		for(uint16_t addr = 0; addr < dataSize; addr++)
		{
			pBuf[0] = 0x20;
			pBuf[1] = (addr >> 8);
			pBuf[2] = (addr & 0xFF);
			pBuf[3] = 0x00;

			pBuf[4] = 0x28;
			pBuf[5] = (addr >> 8);
			pBuf[6] = (addr & 0xFF);
			pBuf[7] = 0x00;

			header.length = 8;

			result = SPIPutPacket(pAT->pSPI, &header, 8, pBuf);
			if(result)
				return result;

			if(waitForData(pAT, 2000))
				return ERROR_ATMEGA_TIMEOUT;

			if(pAT->rx[3] != pData[addr*2] || pAT->rx[7] != pData[addr*2+1])
				return ERROR_ATMEGA_VERIFICATION_FAILED;
		}
	}

	return 0;
}

int16_t ATMegaUpdateProgram(ATMega* pAT, const uint8_t* pData, uint16_t dataSize, uint8_t* pUpdated)
{
	if(pUpdated)
		*pUpdated = 0;

	int16_t result = ATMegaEnterProgramming(pAT);
	if(result == 0)
	{
		result = ATMegaReadSignature(pAT);
		if(result == 0)
		{
			result = ATMegaReadFuses(pAT);
			if(result == 0)
			{
				result = ATMegaVerifyProgram(pAT, pData, dataSize);
				if(result == ERROR_ATMEGA_VERIFICATION_FAILED)
				{
					result = ATMegaChipErase(pAT);
					if(result == 0)
					{
						result = ATMegaLoadProgram(pAT, pData, dataSize);
						if(result == 0)
						{
							if(pUpdated)
								*pUpdated = 1;
							result = ATMegaVerifyProgram(pAT, pData, dataSize);
						}
					}
				}
				else if(result != 0)
				{
					ATMegaExitProgramming(pAT);

					return result;
				}
			}
		}
	}
	else
		return result;

	ATMegaExitProgramming(pAT);

	return result;
}
