#include "pid.h"

void PIDUpdate(PID* pPid)
{
	//max safe feedback and output range: -10000 to 10000

	int32_t current = (*pPid->funcs.feedback)();

	pPid->lastErrorT2 = pPid->lastErrorT1;
	pPid->lastErrorT1 = pPid->lastErrorT0;
	pPid->lastErrorT0 = pPid->setpoint - current;

	int32_t pError = pPid->lastErrorT0 - pPid->lastErrorT1;
	int32_t iError = pPid->lastErrorT1;
	int32_t dError = pPid->lastErrorT0 - 2*pPid->lastErrorT1 + pPid->lastErrorT2;

	if(pPid->frictionKick != 0 && current == 0 && pPid->setpoint != 0)
		pPid->lastOutput = pPid->frictionKick;
	else
		pPid->lastOutput = pPid->lastOutput + (pPid->Kp * pError + pPid->Ki * iError + pPid->Kd * dError)/10000;

	// prevent large output steps
	if((pPid->lastOutput - pPid->log.output) > pPid->slewMax)
		pPid->lastOutput = pPid->log.output + pPid->slewMax;
	else if((pPid->log.output - pPid->lastOutput) > pPid->slewMax)
		pPid->lastOutput = pPid->log.output - pPid->slewMax;

	// limit integral error (prevent runaway during block situation)
/*	if(pPid->integral > pPid->max*20)
		pPid->integral = pPid->max*20;

	if(pPid->integral < -(pPid->max*20))
		pPid->integral = -(pPid->max*20);
*/

	// prevent outputs greater than maximum
	pPid->log.overload = 0;
	if(pPid->lastOutput > pPid->max)
	{
		pPid->lastOutput = pPid->max;
		pPid->log.overload = 1;
	}

	if(pPid->lastOutput < -pPid->max)
	{
		pPid->lastOutput = -pPid->max;
		pPid->log.overload = 1;
	}

	pPid->log.current = current;
	pPid->log.pError = pError;
	pPid->log.iError = iError;
	pPid->log.dError = dError;
	pPid->log.output = pPid->lastOutput;

	(*pPid->funcs.output)(pPid->lastOutput);
}

void PIDInit(PID* pPid, PIDOutputFunc outFunc, PIDFeedbackFunc feedbackFunc)
{
	pPid->Kp = 0;
	pPid->Ki = 0;
	pPid->Kd = 0;

	pPid->lastErrorT0 = 0;
	pPid->lastErrorT1 = 0;
	pPid->lastErrorT2 = 0;
	pPid->lastOutput = 0;
	pPid->setpoint = 0;
	pPid->max = 1;
	pPid->dt = 1;
	pPid->slewMax = 250;
	pPid->frictionKick = 0;

	pPid->log.current = 0;
	pPid->log.pError = 0;
	pPid->log.iError = 0;
	pPid->log.dError = 0;
	pPid->log.output = 0;
	pPid->log.overload = 0;

	pPid->funcs.output = outFunc;
	pPid->funcs.feedback = feedbackFunc;
}

/**
 * Set PID parameters.
 *
 * @param pPid Pointer to PID structure.
 * @param kp Proportional gain [e-4]
 * @param ki Integral gain [e-4]
 * @param kd Differential gain [e-4]
 * @param dt delta T [s*e-4]
 * @param max Estimated maximum (feedback) value.
 */
void PIDSetConfiguration(PID* pPid, int32_t kp, int32_t ki, int32_t kd, int32_t dt, int32_t max, int32_t frictionKick)
{
	pPid->Kp = kp;
	pPid->Ki = ki;
	pPid->Kd = kd;
	pPid->dt = dt;
	pPid->max = max;
	pPid->frictionKick = frictionKick;
}

void PIDSetSetpoint(PID* pPid, int32_t sp)
{
	if(sp > pPid->max)
		sp = pPid->max;

	if(sp < -pPid->max)
		sp = -pPid->max;

	pPid->setpoint = sp;
}

void PIDSetZero(PID* pPid)
{
	pPid->lastErrorT0 = 0;
	pPid->lastErrorT1 = 0;
	pPid->lastErrorT2 = 0;

	pPid->lastOutput = 0;
}
