package edu.dhbw.mannheim.tigers.sim.controller;


import java.util.Enumeration;
import java.util.LinkedHashMap;

import org.apache.commons.configuration.ConfigurationUtils;
import org.apache.log4j.Appender;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.xmi.impl.XMLResourceFactoryImpl;

import edu.dhbw.mannheim.tigers.sim.Sim;
import edu.dhbw.mannheim.tigers.sim.engine.model.IModelDataProvider;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.IPluginProvider;
import edu.dhbw.mannheim.tigers.sim.engine.resources.ITextureProvider;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimResourceException;
import edu.dhbw.mannheim.tigers.sim.gui.IGUI;
import edu.dhbw.mannheim.tigers.sim.gui.swing.SwingGUI;
import edu.dhbw.mannheim.tigers.sim.resources.IImageProvider;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.util.logging.JULLoggingBridge;
import edu.dhbw.mannheim.tigers.sim.util.logging.StartupLogAppender;


/**
 * This class handles the life-cycle of the whole application, and delegates the different aspects to its
 * {@link ASubController}
 * 
 * @author Gero
 */
public class MainController implements IMainController {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	protected Logger												log;
	
	/** Sub controllers */
	private final LinkedHashMap<Type, ASubController>	subs	= new LinkedHashMap<Type, ASubController>();
	
	/** The different types of sub controllers */
	public enum Type {
		INPUT,
		ENGINE;
	}
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	public MainController(String logLevel) {
		StartupLogAppender startupLogger = setup(logLevel);
		initialize(startupLogger);
	}
	

	// --------------------------------------------------------------------------
	// --- initialization and shutdown ------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Should load and initialize logging and all stuff necessary, as textures, models and so on
	 */
	private StartupLogAppender setup(String logLevel) {
		
		// ##### Logging
		// Log4j
		Level level = Level.toLevel(logLevel);
		
		BasicConfigurator.configure(); // Log4j
		final Logger root = Logger.getRootLogger();
		root.setLevel(level);
		
		final Enumeration<?> appenders = root.getAllAppenders();
		while (appenders.hasMoreElements()) {
			Object elem = appenders.nextElement();
			
			if (elem instanceof Appender) {
				Appender appender = (Appender) elem;
				appender.setLayout(new PatternLayout(Sim.LOGGER_LAYOUT));
			}
		}
		
		// java.util.logging
		JULLoggingBridge.install(); // Wraps the Java-Logger (used by jME and jMEPhysics!) to make it
												// use the Log4j back-end
		java.util.logging.Level jLevel = null;
		try {
			if (logLevel != null && !logLevel.isEmpty()) {
				jLevel = java.util.logging.Level.parse(logLevel);
			} else {
				jLevel = java.util.logging.Level.SEVERE;
			}
		} catch (IllegalArgumentException err) {
			jLevel = java.util.logging.Level.SEVERE;
		}
		
		java.util.logging.Logger.getLogger("").setLevel(jLevel);
		
		// Now, suppress annoying debug-messages from Apache Commons Configuration
		Logger.getLogger(ConfigurationUtils.class).setLevel(Level.WARN);
		

		// Instantiate logger for this class
		log = Logger.getLogger(getClass());
		
		// Instantiate a StartupLogAppender for the GUI
		final StartupLogAppender startupLogger = new StartupLogAppender();
		Logger.getRootLogger().addAppender(startupLogger);
		

		// ##### Setup
		log.info("Setting up application...");
		
		// Prepare EMF (for use outside of eclipse RCP)
		Resource.Factory.Registry.INSTANCE.getExtensionToFactoryMap().put("xml", new XMLResourceFactoryImpl());
		
		log.info("Application set up.");
		

		// ##### Resources
		log.info("Loading data...");
		
		try {
			ISettingsProvider.INSTANCE.initialize();
			ITextureProvider.INSTANCE.initialize();
			IImageProvider.INSTANCE.initialize();
			IModelDataProvider.INSTANCE.initialize();
		} catch (SimResourceException exc) {
			log.warn("At least a part of the needed data could not be loaded!", exc);
		}
		
		log.info("Data loaded.");
		

		// ##### Plugins
		log.info("Loading plugins...");
		
		IPluginProvider.INSTANCE.loadPlugins();
		
		log.info("Plugins loaded.");
		

		return startupLogger;
	}
	

	/**
	 * Starts the application by connecting {@link IInputController} with the {@link IGUI}
	 */
	private void initialize(StartupLogAppender startupLogger) {
		
		log.info("Initializing Application...");
		
		// Get Gui...
		IGUI g = new SwingGUI(getInputController());
		
		// ...and hook them up! ;-)
		InputController inc = (InputController) getInputController();
		inc.setGUI(g);
		g.intialize(startupLogger);
		
		log.info("Initialized Application.");
	}
	

	@Override
	public void doExitApp() {
		
		log.info("Preparing application shutdown...");
		
		for (ASubController sub : subs.values()) {
			if (!sub.verifyAppExit()) {
				
				log.info("Application won't shut down, according to: '" + sub + "'");
				return;
			}
		}
		

		// Let SubControllers prepare for coming exit
		for (ASubController sub : subs.values()) {
			sub.prepareAppExit();
		}
		

		// Clean own stuff
		log.info("Unload data...");
		
		ISettingsProvider.INSTANCE.cleanup();
		ITextureProvider.INSTANCE.cleanup();
		IImageProvider.INSTANCE.cleanup();
		IModelDataProvider.INSTANCE.cleanup();
		
		IPluginProvider.INSTANCE.cleanup();
		
		log.info("Data unloaded.");
		

		log.info("Application shutdown prepared.");
		
		shutdown();
	}
	

	/**
	 * The ultimate application shutdown! ;-)
	 */
	private void shutdown() {
		
		log.info("Application shutdown.");
		System.exit(0);
	}
	

	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	public ASubController getController(Type type) {
		
		ASubController result = subs.get(type);
		
		if (result == null) {
			switch (type) {
				
				case ENGINE:
					result = new SimEngineController(this);
					break;
				
				case INPUT:
					result = new InputController(this);
					break;
			}
			
			subs.put(type, result);
		}
		return result;
	}
	

	@Override
	public ISimEngineController getEngineController() {
		
		return (ISimEngineController) getController(Type.ENGINE);
	}
	

	@Override
	public IInputController getInputController() {
		
		return (IInputController) getController(Type.INPUT);
	}
}
