package edu.dhbw.mannheim.tigers.sim.engine.adapter;


import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ExecutionException;

import org.apache.log4j.Logger;

import com.jme.renderer.ColorRGBA;

import edu.dhbw.mannheim.tigers.sim.engine.ICore;
import edu.dhbw.mannheim.tigers.sim.engine.com.ERoboCupTeam;
import edu.dhbw.mannheim.tigers.sim.engine.com.IRoboCupControlProvider;
import edu.dhbw.mannheim.tigers.sim.engine.com.bot.ABotController;
import edu.dhbw.mannheim.tigers.sim.engine.com.bot.BotControllerFactory;
import edu.dhbw.mannheim.tigers.sim.engine.com.cam.SimCamInputProvider;
import edu.dhbw.mannheim.tigers.sim.engine.com.visualization.IVisualizationDataInterpreter;
import edu.dhbw.mannheim.tigers.sim.engine.com.visualization.VisualizationDataReceiver;
import edu.dhbw.mannheim.tigers.sim.engine.model.EModelColorMode;
import edu.dhbw.mannheim.tigers.sim.engine.model.IModelDataProvider;
import edu.dhbw.mannheim.tigers.sim.engine.model.ball.IBallModelControl;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.BotID;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.BotIDProvider;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.IBasicBotModelControl;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.field.IRoboCupFieldData;
import edu.dhbw.mannheim.tigers.sim.engine.modes.match.IMatchModeControl;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimSettingsException;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.IMatchModeProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ITeamProperties;


/**
 * This class should connect the simulation of a RoboCup-SSL-Match with external control-software
 * 
 * @author Gero
 */
public class RoboCupMatchControlAdapter extends ASimEngineAdapter implements IRoboCupControlProvider {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// System
	protected final Logger													log					= Logger.getLogger(getClass());
	
	protected final IMatchModeProperties								props;
	

	// Controls
	private final IMatchModeControl										matchControl;
	
	private final LinkedHashMap<BotID, IBasicBotModelControl>	blueBotControls	= new LinkedHashMap<BotID, IBasicBotModelControl>();
	private final LinkedHashMap<BotID, IBasicBotModelControl>	yellowBotControls	= new LinkedHashMap<BotID, IBasicBotModelControl>();
	private IBallModelControl												ballControl;
	
	private final BotControllerFactory									controllerFactory	= BotControllerFactory.getInstance();
	

	// Input
	private final LinkedList<ABotController<?, ?, ?>>				blueController		= new LinkedList<ABotController<?, ?, ?>>();
	private final LinkedList<ABotController<?, ?, ?>>				yellowController	= new LinkedList<ABotController<?, ?, ?>>();
	
	private VisualizationDataReceiver									visReceiver;
	private IVisualizationDataInterpreter								visInterpreter;
	

	// Output
	private SimCamInputProvider											camProvider;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see RoboCupMatchControlAdapter
	 * @param engine
	 * @param robocupControl
	 * @throws SimSettingsException
	 */
	public RoboCupMatchControlAdapter(ICore engine, IMatchModeControl robocupControl) {
		super("RoboCupModeAdapter", engine);
		
		this.matchControl = robocupControl;
		
		this.props = ISettingsProvider.INSTANCE.getSimProperties().getMatchModeProperties();
		
		core.addObservers(this);
	}
	

	// --------------------------------------------------------------------------
	// --- life cycle ----------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected synchronized void onStart() {
		
		// ##### Add bots and ball to simulation, plus setup Input!
		try {
			
			// Ball
			ballControl = matchControl.addBallToSimulation().get();
			
			EModelColorMode colorMode = EModelColorMode.valueOf(props.getModelColorMode());
			if (colorMode == null) {
				colorMode = EModelColorMode.TEAM;
				log.debug("'" + props.getModelColorMode() + "' is no proper EModelColorMode, took '" + colorMode
						+ "' instead!");
			}
			

			// Bot-Teams
			ITeamProperties propsBlue = props.getTeamBlue();
			final int numberOfBlueBots = propsBlue.getNumberOfBots();
			if (numberOfBlueBots > 0) {
				addTeam(propsBlue.getNumberOfBots(), propsBlue.getFirstBotInputPort(), propsBlue.getBotModelType(),
						blueBotControls, blueController, ERoboCupTeam.BLUE_TEAM, propsBlue.getBotIDs(), colorMode);
			}
			
			ITeamProperties propsYellow = props.getTeamYellow();
			final int numberOfYellowBots = propsYellow.getNumberOfBots();
			if (numberOfYellowBots > 0) {
				addTeam(propsYellow.getNumberOfBots(), propsYellow.getFirstBotInputPort(), propsYellow.getBotModelType(),
						yellowBotControls, yellowController, ERoboCupTeam.YELLOW_TEAM, propsYellow.getBotIDs(), colorMode);
			}
			
		} catch (InterruptedException err) {
			log.warn("Couldn't add bots or ball to the simulation!", err);
		} catch (ExecutionException err) {
			log.warn("Couldn't add bots or ball to the simulation!", err);
		}
		

		// ##### Setup output
		camProvider = new SimCamInputProvider(this);
		camProvider.start();
		

		// ##### Setup visualization input
		visInterpreter = matchControl.getVisualizationDataInterpreter();
		visReceiver = new VisualizationDataReceiver(visInterpreter);
		visReceiver.start();
	}
	

	@Override
	protected synchronized void onResume() {
		setBlocked(false);
	}
	

	private void addTeam(final int numberOfBots, final int firstBotInputPort, final String type,
			final LinkedHashMap<BotID, IBasicBotModelControl> controls,
			final LinkedList<ABotController<?, ?, ?>> controller, ERoboCupTeam team, final String idChain,
			EModelColorMode colorMode) throws InterruptedException, ExecutionException {
		
		int bots = checkNumberOfBot(numberOfBots);
		

		// Read
		List<Integer> ids = readIDs(idChain, bots);
		
		if (ids.size() < bots) {
			log.warn("Less IDs then bots defined, will generate necessary IDs!");
		}
		
		BotIDProvider.getInstance().enqueue(ids, team); // Ask BotIDProvider to use the given IDs
		
		// Add bots to simulation and get their controls
		List<IBasicBotModelControl> list = new ArrayList<IBasicBotModelControl>();
		if (ERoboCupTeam.BLUE_TEAM.equals(team)) {
			ColorRGBA botColor = getBotColor(colorMode, ColorRGBA.blue);
			
			List<IBasicBotModelControl> newControls = matchControl.addBotModelToSimulation(bots, type, botColor, team,
					true).get();
			if (newControls == null) {
				log.debug("Error while adding team...");
				return;
			}
			
			list.addAll(newControls);
		} else {
			ColorRGBA botColor = getBotColor(colorMode, ColorRGBA.yellow);
			
			List<IBasicBotModelControl> newControls = matchControl.addBotModelToSimulation(bots, type, botColor, team,
					false).get();
			if (newControls == null) {
				log.debug("Error while adding team...");
				return;
			}
			
			list.addAll(newControls);
		}
		
		// Input
		int port = firstBotInputPort;
		ABotController<?, ?, ?> tempController = null;
		
		for (IBasicBotModelControl control : list) {
			controls.put(control.getID(), control); // Add controls to HashMap
			
			tempController = controllerFactory.createBotController(control, this, port, team);
			
			if (tempController != null) {
				controller.add(tempController); // Add to list
				tempController.start(); // Start it!
			}
			
			tempController = null;
			
			port++; // Increment port to avoid double port bindings
		}
	}
	

	private int checkNumberOfBot(final int valueFromSettings) {
		int result = valueFromSettings;
		
		if (valueFromSettings < 0) {
			result = 0;
			log.warn("Unsupported number of bots per team (" + valueFromSettings + "), took '" + result + "' instead!");
		} else if (valueFromSettings > 5) {
			result = 5;
			log.warn("Unsupported number of bots per team (" + valueFromSettings + "), took '" + result + "' instead!");
		}
		
		return result;
	}
	

	private ColorRGBA getBotColor(EModelColorMode mode, ColorRGBA color) {
		color = color.clone();
		switch (mode) {
			case DEBUG:
				color.a *= 0.3f;
				return color;
				
			case REALISTIC:
				return color = new ColorRGBA(0.05f, 0.05f, 0.05f, 1);
				
			default:
			case TEAM:
				return color;
		}
	}
	

	/**
	 * Converts a
	 * 
	 * @param idChain (String) of - hopefully - comma-separated int-values of maximum
	 * @param maxIDCount IDs to return
	 * @return a list of {@link Integer}
	 */
	private List<Integer> readIDs(String idChain, int maxIDCount) {
		
		final List<Integer> result = new ArrayList<Integer>();
		
		String[] idStrings = idChain.split(",");
		
		for (int i = 0; i < idStrings.length && i < maxIDCount; i++) {
			try {
				result.add(new Integer(idStrings[i]));
			} catch (NumberFormatException exc) {
				log.warn("Not all given IDs are no legal int values ('" + idStrings[i] + "')!");
			}
		}
		
		return result;
	}
	

	@Override
	protected synchronized void onPause() {
		setBlocked(true);
	}
	

	@Override
	protected synchronized void onStop() {
		// We don't need to implement the cleanup here, as it is already triggered by the end of the
		// simulation loop.
	}
	

	@Override
	public synchronized void cleanup() {
		
		// Close everything
		for (ABotController<?, ?, ?> controller : blueController) {
			controller.stop();
		}
		for (ABotController<?, ?, ?> controller : yellowController) {
			controller.stop();
		}
		
		if (camProvider != null) {
			camProvider.stop();
			camProvider = null;
		}
		
		if (visReceiver != null) {
			visReceiver.stop();
			visReceiver = null;
		}
		
		// Close active observer
		shutdown();
	}
	

	private void setBlocked(boolean blocked) {
		
		// Stop forwarding input
		for (ABotController<?, ?, ?> controller : blueController) {
			controller.setBlocked(blocked);
		}
		for (ABotController<?, ?, ?> controller : yellowController) {
			controller.setBlocked(blocked);
		}
		
		// Stop sending output
		if (camProvider != null) {
			camProvider.setBlocked(blocked);
		}
	}
	

	// --------------------------------------------------------------------------
	// --- IRoboCupWorldControlProvider -----------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public IBallModelControl getBall() {
		return ballControl;
	}
	

	@Override
	public IBasicBotModelControl getBot(BotID id) {
		IBasicBotModelControl botControl = yellowBotControls.get(id);
		
		if (botControl == null) {
			botControl = blueBotControls.get(id);
		}
		
		return botControl;
	}
	

	@Override
	public List<BotID> getBots() {
		ArrayList<BotID> list = new ArrayList<BotID>();
		
		if (blueBotControls != null) {
			list.addAll(blueBotControls.keySet());
		}
		
		if (yellowBotControls != null) {
			list.addAll(yellowBotControls.keySet());
		}
		
		return list;
	}
	

	@Override
	public ERoboCupTeam getTeam(BotID id) {
		IBasicBotModelControl control = null;
		
		control = yellowBotControls.get(id);
		if (control != null) {
			return ERoboCupTeam.YELLOW_TEAM;
		}
		
		control = blueBotControls.get(id);
		if (control != null) {
			return ERoboCupTeam.BLUE_TEAM;
		}
		return null;
	}
	

	@Override
	public List<BotID> getTeamIDs(ERoboCupTeam team) {
		ArrayList<BotID> list = new ArrayList<BotID>();
		
		switch (team) {
			case BLUE_TEAM:
				list.addAll(blueBotControls.keySet());
				break;
			
			default:
			case YELLOW_TEAM:
				list.addAll(yellowBotControls.keySet());
				break;
		}
		
		return list;
	}
	
	
	@Override
	public IRoboCupFieldData getFieldData() {
		final String fieldType = props.getFieldModelType();
		return IModelDataProvider.INSTANCE.getFieldData(fieldType);
	}
}
