package edu.dhbw.mannheim.tigers.sim.engine.com.bot;


import java.io.IOException;

import org.apache.log4j.Logger;


/**
 * The first part of the "botcom-trio": Receives commands of the given type
 * 
 * @see ABotCommunicator
 * @see ABotStatusTransmitter
 * 
 * 
 * @author Gero
 * 
 * @param <C> The base class for the command type
 */
public abstract class ABotCommandReceiver<C> implements Runnable {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final Logger					log						= Logger.getLogger(getClass());
	

	protected final ICommunicator<C>	comm;
	private Thread							thread;
	
	private boolean						expectedIOException	= false;
	
	private final String					name;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see ABotCommandReceiver
	 * @param communicator
	 * @param inputStream
	 * @param name
	 */
	public ABotCommandReceiver(ICommunicator<C> communicator, String name) {
		this.comm = communicator;
		this.name = name;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Starts the thread and opens the connection
	 */
	public void start() {
		
		thread = new Thread(this, "BotCmdReceiver " + name);
		thread.start();
	}
	

	@Override
	public void run() {
		log.debug("Receiver started!");
		
		while (!Thread.currentThread().isInterrupted()) {
			try {
				
				interpret();
				
			} catch (IOException e) {
				if (!expectedIOException) {
					log.warn("Connection error; closing connection...");
				}
				
				expectedIOException = false;
				break;
			}
		}
		
		stop();
	}
	

	/**
	 * This method is called in a loop, which is running as long as the {@link ABotCommunicator} is
	 * on. Identified messages can be passed by using {@link #comm} calling {@link ICommunicator#newCommand(Object)}
	 * 
	 * @throws IOException
	 */
	protected abstract void interpret() throws IOException;
	

	/**
	 * Stops the thread and closes the input stream of this receiver
	 */
	public void stop() {
		expectedIOException = true;
		
		if (thread != null) {
			thread.interrupt();
		}
		
		try {
			closeInput();
		} catch (IOException e) {
			log.error("Error while closing command-receivers InputStream!", e);
		}
		
		comm.notifyShutdown();
	}
	
	
	/**
	 * Closes the input
	 * 
	 * @throws IOException
	 */
	protected abstract void closeInput() throws IOException;
}
