package edu.dhbw.mannheim.tigers.sim.engine.com.bot.ct;


import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.engine.com.bot.ABotCommunicator;
import edu.dhbw.mannheim.tigers.sim.engine.com.bot.IBotCommandTransmitter;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.ACommand;


/**
 * Communication unit of every @CTBotInputInterpreter.
 * It has a @CTBotCommandTransmitter and a @CTBotCommandReceiver
 * which are running in own Threads and are responsible
 * for the communication to the central Software.
 * 
 * Transmitter: Sends Bot-Status Messages.
 * Receiver: Gets Bot-Commands. (Are passed to the {@link CTBotController})
 * 
 * @author Gero, Malte
 */
public class CTBotCommunicator extends ABotCommunicator<ACommand, ACommand> {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final Logger					log	= Logger.getLogger(getClass());

	/** [ms] As this is not used by the CT-Bot, this is quite big */
	private static final int			TIMEOUT = 5000;
	
	private int								port;
	private ServerSocket					server;
	private Socket							clientSocket;
	
	private CTBotStatusTransmitter	transmitter;
	private CTBotCommandReceiver		receiver;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see CTBotCommunicator
	 * @param port
	 * @param botController
	 */
	public CTBotCommunicator(int port, ICTBotController botController) {
		super(botController, TIMEOUT);
		
		this.port = port;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void initialize() {
		// Initialize
		while (server == null) {
			try {
				server = new ServerSocket(port);
			} catch (IOException e) {
				log.debug("Couldn't instantiate a ServerSocket on port: " + port);
				port++;
			}
		}
		log.info("Communicator initialized on port: " + port);
	}
	

	@Override
	protected boolean startTransceiving() throws IOException {
		clientSocket = server.accept();
		
		receiver = new CTBotCommandReceiver(this, clientSocket.getInputStream(), botController.getID().toString());
		receiver.start();
		
		transmitter = new CTBotStatusTransmitter(this, clientSocket.getOutputStream(),
				(ICTBotStatusProvider) botController);
		transmitter.start();
		
		return true;
	}
	
	
	@Override
	public void send(ACommand msg) throws IOException
	{
		transmitter.send(msg);
	}
	

	@Override
	protected void stopTransceiving() {
		if (transmitter != null) {
			transmitter.stop();
		}
		if (receiver != null) {
			receiver.stop();
		}
		
		try {
			if (clientSocket != null) {
				clientSocket.close();
				clientSocket = null;
			}
			
		} catch (IOException err) {
			log.error("Error while closing client socket!", err);
		}
	}
	

	@Override
	protected void deinitialize() {
		try {
			if (server != null) {
				server.close();
				server = null;
			}
		} catch (IOException err) {
			log.error("Error while closing ServerSocket!", err);
		}
	}


	@Override
	protected IBotCommandTransmitter<ACommand> getTransmitter() {
		return transmitter;
	}
}
