package edu.dhbw.mannheim.tigers.sim.engine.com.visualization;


import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import com.google.protobuf.InvalidProtocolBufferException;

import edu.dhbw.mannheim.tigers.sim.engine.com.IStartableInterface;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.visualization.SimVisMessages.SimVisData;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.IMatchModeProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ISimProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ITeamProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.IVisualizationProperties;
import edu.dhbw.mannheim.tigers.sim.util.collection.Pair;
import edu.dhbw.mannheim.tigers.sim.util.network.IReceiver;
import edu.dhbw.mannheim.tigers.sim.util.network.MulticastUDPReceiver;
import edu.dhbw.mannheim.tigers.sim.util.network.NetworkUtility;
import edu.dhbw.mannheim.tigers.sim.util.network.UnicastUDPReceiver;


/**
 * This class is meant to receive incoming {@link VisualizationData}-packets.<br/>
 * <br/>
 * WARNING: The Thread is blocking, but without any brake. If there's enough data incoming, it'll
 * try to consume it!
 * 
 * @author Gero
 */
public class VisualizationDataReceiver implements IStartableInterface, Runnable, IReceiver {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	protected final Logger														log			= Logger.getLogger(getClass());
	
	// Connection
	private IReceiver																receiver;
	private IVisualizationProperties											props;
	
	private boolean																multicastMode;
	private int																		port;
	private final List<Pair<InetSocketAddress, NetworkInterface>>	pairs;
	private final NetworkUtility												networkUtil	= new NetworkUtility();
	
	private final int																BUFFER_SIZE	= 10000;
	private final byte[]															bufferArr	= new byte[BUFFER_SIZE];
	
	private Thread																	thread;
	private volatile boolean													expectIO		= false;
	private final IVisualizationDataInterpreter							interpreter;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see VisualizationDataReceiver
	 * @param interpreter
	 */
	public VisualizationDataReceiver(IVisualizationDataInterpreter interpreter) {
		this.interpreter = interpreter;
		
		ISimProperties simProps = ISettingsProvider.INSTANCE.getSimProperties();
		multicastMode = simProps.isMulticastMode();
		props = simProps.getVisualizationProperties();
		
		port = props.getPort();
		

		// Choose network interface
		pairs = new ArrayList<Pair<InetSocketAddress, NetworkInterface>>();
		
		IMatchModeProperties matchProps = ISettingsProvider.INSTANCE.getSimProperties().getMatchModeProperties();
		ITeamProperties blueTeam = matchProps.getTeamBlue();
		ITeamProperties yellowTeam = matchProps.getTeamYellow();
		
		networkUtil.addNifByAddress(blueTeam.getMcastInterface(), props.getAddress(),
				props.getPort(), pairs);
		networkUtil.addNifByAddress(yellowTeam.getMcastInterface(), props.getAddress(),
				props.getPort(), pairs);
		
		// Log
		log.info("Visualizer listens on nif:");
		for (Pair<InetSocketAddress, NetworkInterface> p : pairs) {
			log.info("   " + p.right.getDisplayName());
		}
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void start() {
		if (multicastMode) {
			receiver = new MulticastUDPReceiver(port, pairs);
		} else {
			receiver = new UnicastUDPReceiver(port);
		}
		
		expectIO = false;
		
		thread = new Thread(this, "VisDataReceiver");
		thread.start();
		
		log.info("VisDataReceiver started.");
	}
	

	@Override
	public void run() {
		// Create new buffer
		ByteBuffer buffer = ByteBuffer.wrap(bufferArr);
		
		while (!Thread.currentThread().isInterrupted()) {
			try {
				
				buffer.clear();
				final DatagramPacket packet = new DatagramPacket(buffer.array(), buffer.capacity());
				
				// Wait and receive message into buffer
				receiver.receive(packet);
				
				// Convert buffer to SimVisData
				final ByteArrayInputStream packetIn = new ByteArrayInputStream(packet.getData(), 0, packet.getLength());
				final SimVisData visData = SimVisData.parseFrom(packetIn);
				
				// Interpret incoming data! =)
				interpreter.receive(visData);
			} catch (InvalidProtocolBufferException err) {
				log.debug("Error while parsing message, dropped it!");
				continue;
			} catch (IOException err) {
				if (!expectIO) {
					log.error("Error while receiving visualization data, stopping.", err);
					break;
				}
			}
		}
	}
	

	@Override
	public DatagramPacket receive(DatagramPacket store) throws IOException {
		return receiver.receive(store);
	}
	

	@Override
	public void stop() {
		expectIO = true;
		thread.interrupt();
		
		log.debug("VisDataReceiver stopped!");
	}
	

	@Override
	public void cleanup() throws IOException {
		stop();
	}
	

	@Override
	public int getLocalPort() {
		return port;
	}
	

	@Override
	public InetAddress getLocalAddress() {
		return receiver.getLocalAddress();
	}
	

	@Override
	public boolean isReady() {
		return true;
	}
}
