package edu.dhbw.mannheim.tigers.sim.engine.model.bot;


import java.util.LinkedList;

import org.apache.log4j.Logger;

import com.jme.math.Quaternion;
import com.jme.math.Vector3f;
import com.jme.scene.Node;
import com.jme.system.DisplaySystem;
import com.jmex.physics.CollisionGroup;
import com.jmex.physics.DynamicPhysicsNode;
import com.jmex.physics.PhysicsNode;
import com.jmex.physics.impl.ode.OdeCollisionGroup;
import com.jmex.physics.impl.ode.OdePhysicsSpace;

import edu.dhbw.mannheim.tigers.sim.engine.com.ERoboCupTeam;
import edu.dhbw.mannheim.tigers.sim.engine.model.ASimNode;
import edu.dhbw.mannheim.tigers.sim.engine.model.IUpdateableModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.ModelHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.TimedModelLocation;
import edu.dhbw.mannheim.tigers.sim.engine.model.UserData;
import edu.dhbw.mannheim.tigers.sim.engine.model.UserData.Value;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.observer.BotLocation;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.observer.BotModelObservable;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.observer.IBotModelObservable;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.observer.IBotModelObserver;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.ABotModelPart;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.IBotModelPart;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelException;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelResourceException;


/**
 * The base class for all BotModel implementation providing a {@link IBotModelObservable}, a {@link BotID}, handling
 * {@link CollisionGroup}s, implementing {@link IBotPartsOwner} and some other stuff necessary. =)
 * 
 * @author Gero
 */
public abstract class ABotModel extends ASimNode implements IBasicBotModelControl, IUpdateableModel, IBotPartsOwner {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long							serialVersionUID	= 5996437583838660664L;
	
	protected final Logger								log					= Logger.getLogger(getClass());
	
	// Bot
	private final BotID									id;
	private final EBotModelType						botType;
	private final IBotModelObservable				observable;
	private final BotLocation							location;
	private String									botInformation;
	
	private final LinkedList<IBotModelPart>		parts					= new LinkedList<IBotModelPart>();
	
	// Physic
	private final LinkedList<DynamicPhysicsNode>	physicNodes			= new LinkedList<DynamicPhysicsNode>();
	
	// Collision
	private final CollisionGroup						collisionGroup;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see ABotModel
	 * @param space
	 * @param display
	 * @param modelType
	 * @param name
	 * @param team
	 * @throws SimModelResourceException
	 * @throws SimModelException
	 */
	public ABotModel(OdePhysicsSpace space, DisplaySystem display, EBotModelType modelType, String name, ERoboCupTeam team)
			throws SimModelResourceException, SimModelException {
		super(space, display, name);
		
		this.botType = modelType;
		this.id = BotIDProvider.getInstance().getNewBotID(team);
		this.location = new BotLocation(getID().getIntValue(), TimedModelLocation.EMPTY_MODEL_LOCATION.clone());
		this.observable = new BotModelObservable(location);

		// Collision
		this.collisionGroup = new OdeCollisionGroup(space, this + " CollisionGroup");
		this.collisionGroup.collidesWith(collisionGroup, false);
		this.collisionGroup.collidesWith(space.getStaticCollisionGroup(), true);
		this.collisionGroup.collidesWith(space.getDefaultCollisionGroup(), true);
		
		// Make this detecable as ABotModel!
		this.setUserData(UserData.TYPE, new UserData(Value.BOT));
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Adds given node to the bots {@link CollisionGroup} and disables it!
	 * 
	 * @param phNode
	 */
	public void addPhysicNode(DynamicPhysicsNode phNode) {
		physicNodes.add(phNode);
		phNode.setCollisionGroup(collisionGroup);
		phNode.setActive(false);
	}
	

	@Override
	public void addModelPart(IBotModelPart part) {
		parts.add(part);
	}
	

	/**
	 * Adds the given {@link ABotModelPart} to the {@link ABotModel}s list
	 */
	protected void addBotModelPart(ABotModelPart part) {
		parts.add(part);
		for (DynamicPhysicsNode node : part.getPhysicNodes()) {
			node.setCollisionGroup(collisionGroup);
		}
	}
	

	@Override
	public void attachNode(Node node) {
		this.attachChild(node);
	}
	

	/**
	 * Needs to be called with the 'main' {@link DynamicPhysicsNode} as parameter to make it pickable!
	 * 
	 * @param node
	 */
	protected void makePickable(DynamicPhysicsNode node) {
		ModelHelper.makePickable(node, Value.BOT);
	}
	

	/*
	 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	 * IUpdateableModel
	 */
	@Override
	public void update(long cycleNumber, float tpc) {
		
		for (IBotModelPart part : parts) {
			part.doUpdate(cycleNumber, tpc);
		}
		
		doUpdate(cycleNumber, tpc);
	}
	

	protected abstract void doUpdate(long cycleNumber, float tpc);
	

	/*
	 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	 * IBotModelControl
	 */
	@Override
	public void setLocation(Vector3f newPosition) {
		setLocation(newPosition, true, true, true);
	}
	

	@Override
	public void setLocation(Vector3f newPosition, boolean clearForce, boolean clearTorque, boolean clearVelocities) {
		
		if (clearForce) {
			this.clearForce();
		}
		
		if (clearTorque) {
			this.clearTorque();
		}
		
		if (clearVelocities) {
			this.clearVelocity();
		}
		
		this.setLocalTranslation(newPosition);
	}
	

	@Override
	public void setRotation(Quaternion quaternion) {
		setRotation(quaternion, true, true, true);
	}
	

	@Override
	public void setRotation(Quaternion quaternion, boolean clearForce, boolean clearTorque, boolean clearVelocities) {
		
		if (clearForce) {
			this.clearForce();
		}
		
		if (clearTorque) {
			this.clearTorque();
		}
		
		if (clearVelocities) {
			this.clearVelocity();
		}
		
		this.setLocalRotation(quaternion);
	}
	

	/**
	 * Used to clear any forces acting on the BotModel and all its children
	 */
	public void clearForce() {
		for (DynamicPhysicsNode node : physicNodes) {
			node.clearForce();
		}
	}
	

	/**
	 * Used to clear the Torque on the BotModel and all its children
	 */
	public void clearTorque() {
		for (DynamicPhysicsNode node : physicNodes) {
			node.clearTorque();
		}
	}
	

	/**
	 * Used to clear any velocity on the BotModel and all its children
	 */
	public void clearVelocity() {
		for (DynamicPhysicsNode node : physicNodes) {
			node.setLinearVelocity(Vector3f.ZERO);
			node.setAngularVelocity(Vector3f.ZERO);
		}
	}
	

	/**
	 * Sets the physical parts of the {@link ABotModel} to 'active' using {@link PhysicsNode#setActive(boolean)}
	 * 
	 * @param active
	 * @return If at least one has been changed
	 */
	public boolean setActive(boolean active) {
		
		boolean changed = false;
		for (DynamicPhysicsNode node : physicNodes) {
			changed |= node.setActive(active);
		}
		return changed;
	}
	

	/*
	 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	 * IBotModelObservable
	 */
	@Override
	public boolean addObservers(IBotModelObserver... observers) {
		return observable.addObservers(observers);
	}
	

	@Override
	public void notifyObservers(BotLocation event) {
		observable.notifyObservers(event);
	}
	

	@Override
	public boolean removeObserver(IBotModelObserver observer) {
		return observable.removeObserver(observer);
	}
	

	@Override
	public void notifyObservers() {
		observable.notifyObservers();
	}
	

	protected BotLocation getCurrentBotStatus(Vector3f worldTranslation, Quaternion worldOrientation) {
		location.set(worldTranslation, worldOrientation, System.nanoTime());
		return location;
	}
	
	
	@Override
	public String toString() {
		return botType + " " + getID();
	}
	

	/*
	 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	 * Getter/Setter
	 */
	@Override
	public BotID getID() {
		return id;
	}
	
	@Override
	public void setInfoText(String info) {
		this.botInformation = info;
	}
	
	@Override
	public String getInfoText() {
		return this.botInformation == null ? "" : this.botInformation;
	}
	

	public EBotModelType getBotType() {
		return botType;
	}
	

	public CollisionGroup getCollisionGroup() {
		return collisionGroup;
	}
}
