package edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts;


import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.data;
import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.mm;

import java.util.ArrayList;
import java.util.List;

import com.jme.bounding.BoundingBox;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.scene.shape.Box;
import com.jme.system.DisplaySystem;
import com.jmex.physics.DynamicPhysicsNode;
import com.jmex.physics.Joint;
import com.jmex.physics.TranslationalJointAxis;
import com.jmex.physics.geometry.PhysicsBox;

import edu.dhbw.mannheim.tigers.sim.engine.model.ModelHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.IBotPartsOwner;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.tiger.TigerBotModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.material.RoboCupMaterial;


/**
 * This class represents the plunger, a kicking-{@link ABotModelPart} of the {@link TigerBotModel}
 * 
 * @author Gero
 */
public class Plunger extends ABotModelPart {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// Physicals
	private final DynamicPhysicsNode			plunger;
	private final TranslationalJointAxis	plungerSlide;
	
	// Owner
	private final IKickerOwner					kOwner;
	
	// TODO kicking-data
	private static final float					PLUNGER_HEIGHT				= 5f;
	private static final float					PLUNGER_WIDTH				= 50f;
	private static final float					PLUNGER_LENGTH				= 30f;
	private static final float					PLUNGER_MASS				= 0.050f;		// kg, Catia, 10.2010

	private static final float					PLUNGER_MAX_EXCESS		= mm(3 + 6.5f);			// Measured, 04.11
	
	private static final float					BASE_PLATE_DEPTH			= mm(4);
	private static final float					PLUNGER_OVER_BOTTOM		= BASE_PLATE_DEPTH
																								+ data(PLUNGER_HEIGHT / 2)
																								+ mm(7);		// Measured, 04.11
	private static final float					PLUNGER_FROM_CENTER		= mm(66f - 6.5f - (PLUNGER_LENGTH / 2));	// Measured, 04.11
	

	private static final float					SHOOT_ACC					= 100f;
	private static final float					SHOOT_VEL					= 0.0203f; //empirical tested by daniel, olli, malte
	
	
	// Kick
	private final SimpleCapModel				cap;
	
	/** [s] */
	private static final float	MAX_TIME_TO_RESET		= 0.5f;
	/** Needed to give the lobber enough time to kick before resetting it */
	private volatile float		timeToReset				= 0;
	
	private static final float	ALLOWED_EPSILON		= mm(0.5f);
	private static final float	PLUNGER_RESET_SPEED	= 5f;
	
	private volatile KickMode	currentMode				= KickMode.NONE;
	
	/** [%] */
	private volatile int			kickForce				= 0;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * {@link Plunger}
	 * 
	 * @param owner
	 * @param display
	 * @param body
	 * @param bodyToGroundDist
	 * @param bodyRadius
	 * @param com
	 */
	public Plunger(IBotPartsOwner owner, IKickerOwner kOwner, DisplaySystem display, DynamicPhysicsNode body,
			float bodyToGroundDist, float bodyRadius, Vector3f com) {
		super(owner, display, body, com);
		
		this.kOwner = kOwner;
		
		this.cap = kOwner.getCapModel();
		
		plunger = space.createDynamicNode();
		
		// Visual
		final Vector3f max = new Vector3f(PLUNGER_LENGTH / 2, PLUNGER_HEIGHT / 2, PLUNGER_WIDTH / 2);
		final Vector3f min = max.negate();
		final Box vPlunger = new Box("vPlunger", min, max);
		vPlunger.setModelBound(new BoundingBox());
		vPlunger.updateModelBound();
		
		ModelHelper.color(vPlunger, display, ColorRGBA.lightGray.clone());
		
		plunger.attachChild(vPlunger);
		
		// Physics
		final PhysicsBox phPlunger = plunger.createBox("phPlunger");
		phPlunger.setMaterial(RoboCupMaterial.BOT_BODY);
		phPlunger.setLocalScale(new Vector3f(PLUNGER_LENGTH, PLUNGER_HEIGHT, PLUNGER_WIDTH));
		plunger.setMass(PLUNGER_MASS);
		
		// Position
		// plunger.setLocalRotation(ModelHelper.y90.clone());
		plunger.setLocalTranslation(PLUNGER_FROM_CENTER, data(bodyToGroundDist) + PLUNGER_OVER_BOTTOM, 0);
		plunger.getLocalTranslation().subtractLocal(com);	// Offset
		owner.attachNode(plunger);
		
		// Attach it to the bot
		final Joint plungerSlider = space.createJoint();
		plungerSlide = plungerSlider.createTranslationalAxis();
		plungerSlide.setDirection(Vector3f.UNIT_X.clone());
		
		plungerSlide.setPositionMinimum(0f);
		plungerSlide.setPositionMaximum(PLUNGER_MAX_EXCESS);
		plungerSlide.setAvailableAcceleration(SHOOT_ACC);
		
		plungerSlider.attach(plunger, body);
	}


	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void doUpdate(long cycleNumber, float tpc) {
		
		switch (currentMode) {
			
			case ARM:
				if (kickForce > 0 && cap.getCurrentLoad() > 0 && kOwner.isBallDetected()) {
					// Ball's there, go for it!
					doKick();
				} else {
					doResetKicker(tpc);
				}
				break;
			
			case FORCE:
				if (kickForce > 0 && cap.getCurrentLoad() > 0) {
					// Kick!
					doKick();
				} else {
					// Reset plunger
					doResetKicker(tpc);
				}
				break;
			
			case DISARM:
			default:
				doResetKicker(tpc);
				break;
		}
	}
	

	private void doKick() {
		plungerSlide.setDesiredVelocity(kickForce * cap.getLoadLevel() * SHOOT_VEL);
		cap.completeDischarge();
		kickForce = 0;
		timeToReset = MAX_TIME_TO_RESET;
		currentMode = KickMode.NONE;
	}
	

	private void doResetKicker(float tpc) {
		
		if (timeToReset > 0) {
			timeToReset -= tpc;
		}
		if (timeToReset <= 0) {
			if (plungerSlide.getPosition() >= plungerSlide.getPositionMaximum() - ALLOWED_EPSILON) {
				plungerSlide.setDesiredVelocity(-PLUNGER_RESET_SPEED);
			} else if (plungerSlide.getPosition() <= plungerSlide.getPositionMinimum()
					+ ALLOWED_EPSILON) {
				plungerSlide.setDesiredVelocity(0f);
			}
		}
	}
	

	/**
	 * Fires the plunger HARD
	 * 
	 * @param percent
	 * @param mode
	 */
	public void fire(int percent, KickMode mode) {
		
		// Normalize percent
		if (percent < 0) {
			percent = 0;
		} else if (percent > 100) {
			percent = 100;
		}
		
		if (timeToReset <= 0) {	// Kicker ready
			kickForce = percent;
			currentMode = mode;
		}
		
		// if (mode == KickMode.DISARM || mode == KickMode.ARM) {
		// chargeTime = 0;
		// }
	}
	

	/**
	 * @return kickers current charge [V]
	 */
	public int getCurrentCharge() {
		return Math.round(cap.getCurrentLoad());
	}
	
	
	/**
	 * @return {@link #cap#isCapLoading()}
	 */
	public boolean isCapLoading()
	{
		return cap.isLoading();
	}


	@Override
	public List<DynamicPhysicsNode> getPhysicNodes() {
		final ArrayList<DynamicPhysicsNode> result = new ArrayList<DynamicPhysicsNode>();
		result.add(plunger);
		return result;
	}
}
