package edu.dhbw.mannheim.tigers.sim.engine.model.bot.tiger;

import java.util.Arrays;

import org.apache.log4j.Logger;

import com.jme.math.FastMath;
import com.jmex.physics.RotationalJointAxis;

import edu.dhbw.mannheim.tigers.sim.engine.com.bot.tiger.TigerBotController;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.BotID;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.IBotModelPart;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerMotorSetParams;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerMotorSetParams.MotorMode;


/**
 * This class has been extracted from {@link TigerBotModel} to reduce complexity. It controls the robots 4 axis-motors.
 * When new commands arrive input is given from the {@link TigerBotController} via {@link ITigerMotorControl}.
 * 
 * @author Gero
 */
public class TigerMotorController implements IBotModelPart, ITigerMotorControl {
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final Logger						log					= Logger.getLogger(getClass());
	
	private static final MotorSpeed		ZERO_SPEED			= new MotorSpeed(0, 0, 0, 0);
	
	private final BotID						id;
	/** Used to translate the bot-vector to wheel-rotations [mm] */
	private final float						wheelRadius;
	

	// ##### Vector-conversion
	// Pre-calculated values
	private static final float				ALPHA					= 28 * FastMath.DEG_TO_RAD;
	private static final float				BETA					= 45 * FastMath.DEG_TO_RAD;
	
	private static final float				VXN					= 2 * (FastMath.sin(ALPHA) + FastMath.sin(BETA));
	
	private static final float				SIN_A					= FastMath.sin(ALPHA);
	private static final float				SIN_B					= FastMath.sin(BETA);
	private static final float				COS_A					= FastMath.cos(ALPHA);
	private static final float				COS_B					= FastMath.cos(BETA);
	
	/**
	 * Controller-to-bot conversion of wheel-speed ({@value}) and compensation
	 * of different wheel rotation direction
	 */
	private static final float				C2B					= -1000;
	
	// These factors are the numerical versions from the pseudo-inverse of the matrix represented by the formulas above
	private static final float				aaa					= 0.2763f;
	private static final float				bbb					= 0.34501f;
	private static final float				ccc					= 0.1995f;
	private static final float				ddd					= 0.30049f;
	

	// ##### Motors
	/** [Nm] (real motor torque * Transmission * absolutely arbitrary value) */
	private static final float				MOTOR_TORQUE		= 5.0f;//0.059f;	// * 3.3f;
//	private static final float				MOTOR_TORQUE_STOP	= MOTOR_TORQUE * 0.5f;
	
	/** 1 front right, 2 rear right, 3 front left, 4 rear left */
	private final RotationalJointAxis[]	m						= new RotationalJointAxis[4];
	
	/**
	 * Helps transferring the velocity (properly synchronized) from the wheels joint-axis to {@link #getBotStatus()}.
	 * Holds the speed as "Bahngeschwindigkeit" [mm/s].
	 */
	private volatile MotorSpeed			currentSpeed		= ZERO_SPEED;
	/**
	 * The speed set by the {@link #setSpeedAutomatic(float, float, float)}. Needed to dispatch calls from the outside
	 * from the inner sim-loop ({@link #doUpdate(long, float)}). Holds the speed as "rotations/s".
	 */
	private volatile MotorSpeed			desiredSpeed		= ZERO_SPEED;
	

	// ### Faking the 16-shaped wheels
//	/** Angular distance between two lateral wheels [rad] */
//	private static final float				THRESHOLD_MIN		= 360 / (15 * 2) * FastMath.DEG_TO_RAD;
	

	// ##### Motor logic and controllers
	// Motor state
	private MotorMode							currentMotorMode	= MotorMode.AUTOMATIC;

	private static final long				MOVE_VM2PID			= 104540L;	// [e-6], 99310 @ r_R=3cm
	
	private static final int				PID_MAX				= 9360;
	/** 374 * 15V = {@value}[rps] */
	private static final int				MAX_MOTOR_ROT		= 5610/ 60;
	/** {@link #MAX_MOTOR_ROT} * 0.3f = {@value} */
	private static final float				MAX_WHEEL_ROT		= MAX_MOTOR_ROT * 0.3f;
	private final float						PID_2_MOTOR_JOINT = (2 * FastMath.PI) * MAX_WHEEL_ROT / PID_MAX;
	
	// Controllers
	/** Synced by itself */
	private final MotorPid[]				pids					= new MotorPid[4];
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see TigerMotorController
	 * @param id
	 * @param m1
	 * @param m2
	 * @param m3
	 * @param m4
	 * @param wheelRadius [mm]
	 */
	public TigerMotorController(BotID id, RotationalJointAxis m1, RotationalJointAxis m2, RotationalJointAxis m3,
			RotationalJointAxis m4, float wheelRadius) {
		
		this.id = id;
		this.wheelRadius = wheelRadius;
		
		this.m[0] = m1;
		this.m[1] = m2;
		this.m[2] = m3;
		this.m[3] = m4;
		
		// Set acceleration
		for (RotationalJointAxis motor : m) {
			motor.setAvailableAcceleration(MOTOR_TORQUE);
		}
		
		final int cps = ISettingsProvider.INSTANCE.getSimProperties().getSimEngineProperties().getCPS();
		
		// PIDs
		synchronized (pids) {
			for (int i = 0; i < 4; i++) {
				pids[i] = new MotorPid(m[i], cps);
				// 1000, 1000, 0; ;9000 (from Marvin V3)
				pids[i].configure(1, 0, 0, PID_MAX, PID_MAX);
			}
		}
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Updated by sim-loop
	 */
	@Override
	public void doUpdate(long cycleNumber, float tpc) {
		// Store current speed (for status-generation)
		MotorSpeed newCurrentSpeed = createCurrentMotorSpeed();
		this.currentSpeed = newCurrentSpeed;
		
		// PIDs
		// Setpoint
		MotorSpeed desired = new MotorSpeed(desiredSpeed.v);
		
		
		MotorSpeed controlledDesiredSpeed = null;
		synchronized (pids) {

			for (int i = 0; i < 4; i++) {
				pids[i].updateSetpoint(desired.v[i]);
			}
			
			
			// Update
			for (MotorPid pid : pids) {
				pid.update();
			}
			
			// Retrieve result
			controlledDesiredSpeed = new MotorSpeed(pids[0].getLastOutput() * PID_2_MOTOR_JOINT, pids[1].getLastOutput() * PID_2_MOTOR_JOINT,
					pids[2].getLastOutput() * PID_2_MOTOR_JOINT, pids[3].getLastOutput() * PID_2_MOTOR_JOINT);
		}
		
		desired = controlledDesiredSpeed;
	

		// Set new, formerly dispatched desired speed
		setMotorVelocities(newCurrentSpeed, desired);
	}
	

	/**
	 * Sets the new PID-parameter (called from the {@link TigerBotController} whenever a new command is received)
	 */
	@Override
	public void setPIDParams(TigerMotorSetParams newParams) {
		synchronized (pids) {
			for (int i = 0; i < 4; i++) {
				pids[i].configure(newParams.getKp(i), newParams.getKi(i), newParams.getKd(i), PID_MAX,
						newParams.getSlewMax(i));
			}
		}
	}
	

	/**
	 * As the controller sends the wanted desired velocity, so we have to convert it to the
	 * angular velocity.
	 * 
	 * wheelPerimeter = 2*Pi * r;
	 * v(desired) = (2*Pi) * x / wheelPerimeter;
	 * 
	 * => v(desired) = x / r;
	 */
	@Override
	public void setSpeedAutomatic(float newVx, float newVy, float newVw) {
		if (currentMotorMode != MotorMode.AUTOMATIC) {
			log.warn("Received TigerMotorMove, but bot " + id + " is in " + currentMotorMode + "'! Dropped.");
			return;
		}
		
		// Automatic move
		float vx = newVx; // m/s
		float vy = newVy; // m/s
		float vw = newVw * 0.082f; // [rad/s] * [m] = [m/s]
//		System.out.println("Vx: " + vx + "; Vy: " + vy + "; Vw: " + vw + " (" + newVw);
		final float[] vm = new float[4]; // [m/s]
		
		vm[0] = -SIN_A * vx + COS_A * vy + vw;
		vm[1] = -SIN_A * vx - COS_A * vy + vw;
		vm[2] = +SIN_B * vx - COS_B * vy + vw;
		vm[3] = +SIN_B * vx + COS_B * vy + vw;
		

		// Convert to rotations
		for (int i = 0; i < 4; i++) {
			vm[i] *= C2B; // Convert to [mm/s], regarding axis-oriantations
			vm[i] *= MOVE_VM2PID;
			vm[i] /= 100000;
//			vm[i] /= wheelRadius;	// / 2 * FastMath.PI; // Convert from "Bahngeschwindigkeit" [mm/s] to [rotation/s] [rad/s]
		}
		
		// Set desired speed
		final MotorSpeed newSpeed = new MotorSpeed(vm);
		this.desiredSpeed = newSpeed;
	}
	

	/**
	 * Implements a fake for 16-shape wheels
	 * 
	 * @param desired
	 */
	private void setMotorVelocities(MotorSpeed current, MotorSpeed desired) {
//		if (abs(desired.vel1) < THRESHOLD_MIN && abs(current.vel1) < THRESHOLD_MIN && abs(desired.vel2) < THRESHOLD_MIN
//				&& abs(current.vel2) < THRESHOLD_MIN && abs(desired.vel3) < THRESHOLD_MIN
//				&& abs(current.vel3) < THRESHOLD_MIN && abs(desired.vel4) < THRESHOLD_MIN
//				&& abs(current.vel4) < THRESHOLD_MIN) {
//			setMotorVelocity(m, ZERO_SPEED, MOTOR_TORQUE_STOP);
//		} else {
//			setMotorVelocity(m, desired, MOTOR_TORQUE);	//(m, desired, MOTOR_TORQUE);
			setMotorVelocityDynamic(m, desired, current);	//(m, desired, MOTOR_TORQUE);
//		}
	}
	

//	private float abs(float value) {
//		return FastMath.abs(value);
//	}
	

	@SuppressWarnings("unused")
	private void setMotorVelocity(RotationalJointAxis[] motors, MotorSpeed desired, float torque) {
		if (motors.length != 4) {
			log.error("Invalid motor-array length: " + motors.length + " !!!");
		}
		
		for (int i = 0; i < 4; i++) {
			motors[i].setDesiredVelocity(desired.v[i]);
			motors[i].setAvailableAcceleration(torque);
		}
	}

	private void setMotorVelocityDynamic(RotationalJointAxis[] motors, MotorSpeed desired, MotorSpeed current) {
		if (motors.length != 4) {
			log.error("Invalid motor-array length: " + motors.length + " !!!");
		}
		
		
		for (int i = 0; i < 4; i++) {
			motors[i].setDesiredVelocity(desired.v[i]);
			motors[i].setAvailableAcceleration(calcTorque(current.v[i]));
		}
	}
	
	// Reflects real motor behavior at constant 15V supply voltage and 60mm wheel diameter
	// Includes torque multiplier for transmission
	private float calcTorque(float currentVelocity)
	{
		currentVelocity = FastMath.abs(currentVelocity);	// calculation requires positive values
		
		currentVelocity *= 3.333333f;	// motor turns faster than wheel due to transmission
		float torque = 0.31875f - 0.003409f*(currentVelocity/(FastMath.PI*60.0f));	// Now thats the motor torque
		torque *= 3.333333f;	// and thats the transmission
		
		if(torque < 0)	{// negative torque is not possible
			torque = 0;
		}
		
		return torque;
	}

	/**
	 * Used to store a set of motor-velocities<br/>
	 * <b>NOTE: Values may have individual UNITS!</b>
	 * 
	 * @author Gero
	 */
	private static class MotorSpeed {
		private final float[] v;
		
		
		public MotorSpeed(float vel1, float vel2, float vel3, float vel4) {
			this.v = new float[4];
			
			this.v[0] = vel1;
			this.v[1] = vel2;
			this.v[2] = vel3;
			this.v[3] = vel4;
		}
		
		public MotorSpeed(float[] vels) {
			if (vels.length != 4) {
				throw new IllegalArgumentException("Invalid motorspeed-length: " + vels.length);
			}
			
			this.v = Arrays.copyOf(vels, 4);
		}
	}
	
	
	/**
	 * @return Fills a {@link MotorSpeed} with <code>{@link RotationalJointAxis#getVelocity()} * wheelRadius</code> of
	 *         all motors (Bahngeschwindigkeit)
	 */
	private MotorSpeed createCurrentMotorSpeed() {
		return new MotorSpeed(m[0].getVelocity() * wheelRadius, m[1].getVelocity() * wheelRadius, m[2].getVelocity()
				* wheelRadius, m[3].getVelocity() * wheelRadius);
	}
	

	// public void handleSetPIDSetPoint(TigerMotorSetPidSp pidCmd) {
	// if (currentMotorMode != MotorMode.PID) {
	// log.warn("Received TigerMotorSetPidSp, but bot " + id + " is in " + currentMotorMode + "'! Dropped.");
	// return;
	// }
	// // TODO Manual Motor-values!
	// log.warn("Received TigerMotorSetPidSp, dropped!");
	// }
	//
	//
	// public void handleSetPIDSetManual(TigerMotorSetManual manualCmd) {
	// if (currentMotorMode != MotorMode.MANUAL) {
	// log.warn("Received TigerMotorSetManual, but bot " + id + " is in " + currentMotorMode + "'! Dropped.");
	// return;
	// }
	// // TODO Manual Motor-values!
	// log.warn("Received TigerMotorSetManual, dropped!");
	// }
	

	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @return The current bot-vector in [m/s]
	 */
	TigerMove calcTigerMove() {
		
		// Now we have to adjust to [m/s] and calculate vx, vy and vw (remember: Wheels have negative turn-direction!)
		MotorSpeed sp = currentSpeed;
		float vx = (sp.v[3] / VXN + sp.v[2] / VXN - sp.v[1] / VXN - sp.v[0] / VXN);
		float vy = (sp.v[3] * aaa - sp.v[2] * aaa - sp.v[1] * bbb + sp.v[0] * bbb);
		
		float vw = (sp.v[3] * ccc + sp.v[2] * ccc + sp.v[1] * ddd + sp.v[0] * ddd);
		
		// Convert from motor to
		vx = vx / C2B;
		vy = vy / C2B;
		vw = vw / C2B;
		
		return new TigerMove(vx, vy, vw);
	}
}
