package edu.dhbw.mannheim.tigers.sim.engine.model.field;


import com.jme.bounding.BoundingBox;
import com.jme.math.Quaternion;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.renderer.Renderer;
import com.jme.scene.SharedMesh;
import com.jme.scene.shape.Box;
import com.jme.scene.shape.Cylinder;
import com.jme.scene.state.BlendState;
import com.jme.system.DisplaySystem;
import com.jmex.physics.StaticPhysicsNode;
import com.jmex.physics.impl.ode.OdePhysicsSpace;

import edu.dhbw.mannheim.tigers.sim.engine.model.ASimNode;
import edu.dhbw.mannheim.tigers.sim.engine.model.ModelHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.field.IRoboCupFieldData;
import edu.dhbw.mannheim.tigers.sim.engine.model.material.RoboCupMaterial;
import edu.dhbw.mannheim.tigers.sim.engine.resources.TextureProvider.ETexIndex;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelException;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelResourceException;

/**
 * The programmatical implementation of a RoboCup Goal.
 * 
 * @author Gero
 */
public class Goal extends ASimNode {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long serialVersionUID = -6656407798160223359L;
	
	// Field
	private final float lineWidth;
	
	// Goal
	private final float innerWidth;
	private final float innerHeight;
	private final float innerDepth;
	
	// Top bar
	private final float topDiameter;
	
	// Net
	private final float netWidthSide;
	private final float netHeightSide;
	private final float netWidthBack;

	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	  * @see Goal
	  * @param space
	  * @param display
	  * @param data
	  * @throws SimModelResourceException
	  * @throws SimModelException
	  */
	public Goal(OdePhysicsSpace space,  DisplaySystem display, IRoboCupFieldData data)
			throws SimModelResourceException, SimModelException {
		super(space, display, "goal");
		
		
		
		this.innerWidth = data.getGoalInnerWidth();
		this.innerHeight = data.getGoalInnerHeight();
		this.innerDepth = data.getGoalInnerDepth();
		this.topDiameter = data.getGoalTopbarDiameter();
		this.netWidthSide = data.getGoalSideWallWidth();
		this.netHeightSide = data.getGoalSideWallHeight();
		this.netWidthBack = data.getGoalBackWallWidth();
		this.lineWidth = data.getFieldLineWidth();
		
		createPhysicalNode();
	}

	
	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void createPhysicalNode() throws SimModelResourceException, SimModelException {
		
		/* This will get a bit more complicated:
		 * The goal consist of several parts that may be copied by SharedMeshs to save rendering time.
		 * These are:
		 * goalposts:
		 * - topBar
		 * - goalpost		(copied to the other side)
		 * 
		 * net:
		 * - sideNet		(copied to the other side)
		 * - topNet
		 * - backNet
		 * Their are created around the x-center, y-bottom of the goal and z-behind the goal line
		 * (as if you stand in front of it), so they just have to be moved fielLength along z to their final position.
		 * 
		 * As the Cylinders are created centered, they have to be adjusted in their distance to fit on the line,
		 * and the net must be put a lineWidth towards us to finally sit just on it.=) 
		 * 
		 * From top:
		 * 	---------
		 *    |       |	<- wideWall(/-net)
		 *           
		 *            ^- goalposts
		 */
		
		StaticPhysicsNode goalposts = space.createStaticNode();
		goalposts.setName("goalposts");
		
		// First the top bar				########################################################
		
		// Necessary calculations		
		float topRadius = topDiameter / 2;
		float topWidth = innerWidth + netWidthSide;
		float topCenterHeight = innerHeight + topRadius;
		
		
		Cylinder topBar = new Cylinder("topBar", (int) innerWidth / 10,
				(int) topDiameter * 2, topRadius, topWidth, true);
		topBar.setModelBound(new BoundingBox());

		// Rotate it...
		Quaternion quat = new Quaternion();
		quat.fromAngleNormalAxis(rad(90), Vector3f.UNIT_Z);
		quat.fromAngleNormalAxis(rad(90), Vector3f.UNIT_Y);
		topBar.setLocalRotation(quat);
		topBar.updateModelBound();
		topBar.updateWorldBound();
		
		// And lift it!
		topBar.setLocalTranslation(0, topCenterHeight, topRadius - lineWidth);
		
		// Color and opaqueness
		topBar.setDefaultColor(ColorRGBA.white);
		
		goalposts.attachChild(topBar);
		

		// Then the goalposts			########################################################

		// Necessary calculations
		float goalpostRadius = netWidthSide / 2;
		float goalpostHeight = netHeightSide;
		float goalpostCenterToGoalCenterDistance = (innerWidth / 2) + goalpostRadius; 
		
		
		Cylinder goalpost = new Cylinder("goalpost", (int) innerHeight / 10,
				(int) netWidthSide * 2, goalpostRadius, goalpostHeight, true);
		goalpost.setModelBound(new BoundingBox());

		// Turn them up...
		Quaternion turnupGoalposts = new Quaternion();
		turnupGoalposts.fromAngleAxis(rad(90), Vector3f.UNIT_X);
		
		// Colorize it
		goalpost.setDefaultColor(ColorRGBA.white);
		
		// Use it two times, for each side one:
		SharedMesh leftGoalpost = new SharedMesh("leftGoalpost", goalpost);
		leftGoalpost.setLocalRotation(turnupGoalposts.clone());
		SharedMesh rightGoalpost = new SharedMesh("rightGoalpost", goalpost);
		rightGoalpost.setLocalRotation(turnupGoalposts.clone());
		rightGoalpost.updateModelBound();
		rightGoalpost.updateWorldBound();
		
		// ...move them to left and right, lift 'em out of the ground,
		// and a bit towards us because of their centered creation!
		leftGoalpost.setLocalTranslation(-goalpostCenterToGoalCenterDistance, goalpostHeight / 2, goalpostRadius - lineWidth);
		rightGoalpost.setLocalTranslation(goalpostCenterToGoalCenterDistance, goalpostHeight / 2, goalpostRadius - lineWidth);
		
		goalposts.attachChild(leftGoalpost);
		goalposts.attachChild(rightGoalpost);
		
		
		// Finish goalposts
		goalposts.setRenderQueueMode(Renderer.QUEUE_OPAQUE);
		
		// Physics
		goalposts.setMaterial(RoboCupMaterial.GOALPOST);
		goalposts.generatePhysicsGeometry();
		
		this.attachChild(goalposts);
		
		
		
		// Now the nets at the side	########################################################
		
		// Necessary calculations
		float sideNetCenterToGoalCenterDistance = (innerWidth + netWidthSide)/ 2;
		
		StaticPhysicsNode net = space.createStaticNode();
		net.setName("net");
		
		Box box = new Box("sideNet", new Vector3f(-(netWidthSide / 2), 0, goalpostRadius),
				new Vector3f(netWidthSide / 2, netHeightSide, innerDepth));
		box.setModelBound(new BoundingBox());
		
		SharedMesh leftNet = new SharedMesh("leftNet", box);
		leftNet.setLocalTranslation(-sideNetCenterToGoalCenterDistance, 0, 0);
		leftNet.updateModelBound();
		leftNet.updateWorldBound();
		
		SharedMesh rightNet = new SharedMesh("rightNode", box);
		rightNet.setLocalTranslation(sideNetCenterToGoalCenterDistance, 0, 0);
		rightNet.updateModelBound();
		rightNet.updateWorldBound();

		net.attachChild(rightNet);
		net.attachChild(leftNet);
		
		
		
		// And now the other nets		########################################################
		// Nearly finished. The topNet...
		Box topNet = new Box("topNet",
				new Vector3f(-(topWidth / 2), innerHeight, topRadius),
				new Vector3f(topWidth / 2, innerHeight + topDiameter, innerDepth));

		topNet.setModelBound(new BoundingBox());
		topNet.updateModelBound();
		topNet.updateWorldBound();
		
		net.attachChild(topNet);
		
		// ...and last but not least: the backNet!
		
		// Necessary calculations
		float outerWidth = innerWidth + (netWidthSide * 2);
		
		Box backNet = new Box("backNet",
				new Vector3f(-(outerWidth / 2), 0, 0),
				new Vector3f(outerWidth / 2, netHeightSide, netWidthBack));
		backNet.setModelBound(new BoundingBox());
		
		backNet.setLocalTranslation(0, 0, innerDepth);
		backNet.updateModelBound();
		backNet.updateWorldBound();
		
		net.attachChild(backNet);
		
		// Finish net and make it transparent!=)
		net.setRenderQueueMode(Renderer.QUEUE_TRANSPARENT);
		
		// Texture...
		ModelHelper.applyTexture(net, display, ETexIndex.REFLECTOR);
		
		// ... and the Alphas for real transparency!
		BlendState netAS = display.getRenderer().createBlendState();
		netAS.setBlendEnabled(true);
		netAS.setSourceFunction(BlendState.SourceFunction.SourceAlpha);
		netAS.setDestinationFunction(BlendState.DestinationFunction.One);
		netAS.setTestEnabled(true);
		netAS.setTestFunction(BlendState.TestFunction.GreaterThan);
		netAS.setEnabled(true);
		
		net.setRenderState(netAS);
		
		// Move forward, because the goals final point of reference lies _behind_ the line. 
		net.setLocalTranslation(0, 0, -lineWidth);
		
		// Physic
		net.setMaterial(RoboCupMaterial.NET);
		net.generatePhysicsGeometry();
		
		
		this.attachChild(net);

		
		// Finish goal
//		this.setModelBound(new BoundingBox());
		this.updateModelBound();
	}
}
