package edu.dhbw.mannheim.tigers.sim.engine.modes;

import java.util.LinkedList;

import org.apache.log4j.Logger;

import com.jme.renderer.Renderer;
import com.jme.util.GameTaskQueue;

import edu.dhbw.mannheim.tigers.sim.engine.ASimEngineState;
import edu.dhbw.mannheim.tigers.sim.engine.ESimEngineState;
import edu.dhbw.mannheim.tigers.sim.engine.ICore;
import edu.dhbw.mannheim.tigers.sim.engine.IRenderableSimEnginePart;
import edu.dhbw.mannheim.tigers.sim.engine.ISimEnginePart;
import edu.dhbw.mannheim.tigers.sim.engine.IStatefulSimEngine;
import edu.dhbw.mannheim.tigers.sim.engine.IUpdateableSimEnginePart;
import edu.dhbw.mannheim.tigers.sim.engine.adapter.ASimEngineAdapter;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.ASimDelegate;
import edu.dhbw.mannheim.tigers.sim.engine.model.IUpdateableModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.ModelFactory;


/**
 * <p>
 * Abstract base class for all SimEngineModes. Provides some basic implementations as {@link ISimEnginePart}, and
 * handles its {@link ASimEngineMode}s and {@link ASimEngineAdapter}s.
 * </p>
 * <p>
 * It also uses the idea of jME's {@link GameTaskQueue}.
 * </p>
 * 
 * @author Gero
 */
public abstract class ASimEngineMode implements ISimEnginePart, IRenderableSimEnginePart, IStatefulSimEngine {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// Util
	protected final Logger							log		= Logger.getLogger(getClass());
	protected final ModelFactory					factory;
	

	// System
	protected final ICore							core;
	
	private final ESimEngineMode					mode;
	

	// Delegates
	protected final LinkedList<ASimDelegate>	delegates;
	
	
	protected void addDelegate(ASimDelegate value) {
		delegates.add(value);
	}
	
	
	// Adapters
	protected final LinkedList<ASimEngineAdapter>	adapters;
	
	
	protected void addAdapter(ASimEngineAdapter value) {
		adapters.add(value);
	}
	
	
	// UpdateableModel
	// Doesn't have to be synchronized as models should only be added from the OpenGL-thread! (Use GameTaskQueue!!!)
	protected final LinkedList<IUpdateableModel>	updateableModels;
	
	
	protected boolean addUpdateableModel(IUpdateableModel value) {
		return updateableModels.add(value);
	}
	

	protected boolean removeUpdateableModel(IUpdateableModel value) {
		return updateableModels.remove(value);
	}
	
	// State
	protected ASimEngineState			currentState;
	

	protected final ASimEngineState	pausedFull				= new ASimEngineState(ESimEngineState.PAUSED_FULL) {};
	protected final ASimEngineState	allExceptPausedFull	= new ASimEngineState(
			ESimEngineState.getStatesExceptPausedFull()) {
		@Override
		public void update(float interpolation, float tpc, float cps) {
			for (IUpdateableSimEnginePart delegate : delegates) {
				delegate.update(interpolation, tpc, cps);
			}
		};
	};
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	public ASimEngineMode(ICore core, ESimEngineMode mode) {
		super();
		
		this.core = core;
		this.mode = mode;
		
		this.factory = ModelFactory.getInstance();
		this.factory.setCore(core);
		
		this.delegates = new LinkedList<ASimDelegate>();
		this.adapters = new LinkedList<ASimEngineAdapter>();
		this.updateableModels = new LinkedList<IUpdateableModel>();
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Provides a basic implementation that just differentiates between {@link ESimEngineState#PAUSED_FULL} and
	 * all other states.
	 */
	@Override
	public void changeStateTo(ESimEngineState newState) {
		
		switch (newState) {
			case PAUSED_FULL:
				currentState = pausedFull;
				break;
			
			default:
				currentState = allExceptPausedFull;
				break;
		}
	}
	

	/**
	 * Place for adding new delegates<br/>
	 * Remember: Order is _deadly_ important!!!
	 */
	public abstract void setupDelegates();
	

	/**
	 * Place for adding new adapters to the {@link ASimEngineMode}<br/>
	 * Remember: Order is _deadly_ important!!!<br/>
	 * (Default implementation does nothing, subclasses should override!)
	 */
	public void setupAdapters() {
	};
	

	// --------------------------------------------------------------------------
	// --- ISimEnginePart, IRenderableSimEnginePart and StatefulSimEngine -------
	// --------------------------------------------------------------------------
	/**
	 * Provides a default implementation for the {@link ISimEnginePart#cleanup()} routine which simply calls every
	 * registered delegate.
	 * Try <b>not</b> to override this method but to use the delegate concept instead!
	 */
	@Override
	public void cleanup() {
		
		core.shutdownUpdateQueue();
		
		for (ASimEngineAdapter adapter : adapters) {
			adapter.cleanup();
		}
		
		for (ISimEnginePart delegate : delegates) {
			delegate.cleanup();
		}
	}
	

	/**
	 * Provides a default implementation for the {@link ISimEnginePart#debug(float)} routine which simply calls every
	 * registered delegate
	 * Try <b>not</b> to override this method but to use the delegate concept instead!
	 */
	@Override
	public void debug(float interpolation) {
		
		for (ISimEnginePart delegate : delegates) {
			delegate.debug(interpolation);
		}
	}
	

	/**
	 * Provides a default implementation for the {@link ISimEnginePart#initSimulation()} routine which simply calls every
	 * registered delegate
	 * Try <b>not</b> to override this method but to use the delegate concept instead!
	 */
	@Override
	public void initSimulation() {
		
		for (ISimEnginePart delegate : delegates) {
			delegate.initSimulation();
		}
	}
	

	/**
	 * Provides a default implementation for the {@link ISimEnginePart#initSystem()} routine which simply calls every
	 * registered delegate
	 * Try <b>not</b> to override this method but to use the delegate concept instead!
	 */
	@Override
	public void initSystem() {
		
		setupDelegates();
		
		setupAdapters();
		
		for (ISimEnginePart delegate : delegates) {
			delegate.initSystem();
		}
	}
	

	/**
	 * <p>
	 * Provides a default implementation for the update routine which simply calls every registered delegate as long as
	 * state is not {@link ESimEngineState#PAUSED_FULL}.<br/>
	 * Subclasses should define own states to get more sophisticated results.
	 * </p>
	 * <p>
	 * It also consumes every {@link ITask} from the {@link #updateTaskQueue}.
	 * </p>
	 */
	@Override
	public void update(float interpolation, float tpc, float cps) {
		
		// Update models
		for (IUpdateableModel model : updateableModels) {
			model.update(core.getCurrentCycle(), tpc);
		}
		
		// Update the current state...
		currentState.update(interpolation, tpc, cps);
	}
	

	/**
	 * Provides a default implementation for the {@link IRenderableSimEnginePart#render(float)} routine which does
	 * <b>nothing</b><br/>
	 * Subclasses should override to implement more sophisticated approaches <b>or</b> define own states.
	 */
	@Override
	public void render(float interpolation) {
		// Render the current state
		currentState.render(interpolation);
	}
	
	
	@Override
	public void afterRenderCallback(float interpolation, Renderer renderer) {
		
		// Delegate callback =)
		for (ASimDelegate delegate : delegates) {
			delegate.afterRenderCallback(interpolation, renderer);
		}
	}
	

	@Override
	public ESimEngineState getState() {
		return core.getState();
	}
	

	public ESimEngineMode getMode() {
		return mode;
	}
}
