package edu.dhbw.mannheim.tigers.sim.engine.plugins;


import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.List;

import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.SubnodeConfiguration;
import org.apache.commons.configuration.XMLConfiguration;
import org.apache.commons.configuration.tree.ConfigurationNode;
import org.apache.log4j.Logger;
import org.w3c.dom.DOMConfiguration;
import org.w3c.dom.Document;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSOutput;
import org.w3c.dom.ls.LSSerializer;


/**
 * This class helps the {@link PluginProvider} to handle the plugins config-files and their
 * structure=)
 * 
 * @author Gero
 */
public class PluginConfigHelper {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	protected static final String	XML_ENCODING	= "ISO-8859-1";
	

	private final Logger				log				= Logger.getLogger(getClass());
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	  * @see PluginConfigHelper
	  */
	PluginConfigHelper() {
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	protected XMLConfiguration createAndcheckDefaultsProperties(File file) {
		XMLConfiguration config;
		
		if (!file.exists()) {
			log.debug("No properties-file at '" + file.getPath() + "', creating new one!");
			config = createEmptyConfiguration(file);
		} else {
			try {
				config = new XMLConfiguration(file.getPath());
			} catch (Exception err) {
				log.debug("Unable to read from default properties-file at '" + file.getPath() + "'!",
						err);
				config = createEmptyConfiguration(file);
			}
		}
		
		return config;
	}
	

	private XMLConfiguration createEmptyConfiguration(File f) {
		XMLConfiguration config = new XMLConfiguration();
		config.setRootElementName("plugins");
		config.setFileName(f.getPath());
		config.setAutoSave(true);
		
		writeConfigurationFormatted(config, f);
		
		return config;
	}
	

	protected SubnodeConfiguration getPluginConfigById(XMLConfiguration xml, PluginID id) {
		
		List<?> pluginNodes = xml.configurationsAt("plugin");
		
		for (Object obj : pluginNodes) {
			SubnodeConfiguration config = (SubnodeConfiguration) obj;
			String idStr = config.getString("[@id]", "");
			
			PluginID xmlID = new PluginID(idStr);
			if (id.equals(xmlID)) {
				return config;
			}
		}
		
		// No config yet, create new one
		return createEmptyConfig(xml, id);
	}
	

	private SubnodeConfiguration createEmptyConfig(XMLConfiguration xml, PluginID id) {
		// Add new node at the end using special index '-1'
		xml.addProperty("plugin(-1)[@id]", id.base);
		xml.addProperty("plugin.properties", "");
		
		// Get last, created plugin-node (if no index, last is default!)
		List<?> l = xml.configurationsAt("plugin");
		return (SubnodeConfiguration) l.get(l.size() - 1);
	}
	

	protected void savePluginNodesToFile(List<ConfigurationNode> nodes, File file)
			throws ConfigurationException {
		XMLConfiguration config = new XMLConfiguration();
		config.setRootElementName("plugins");
		config.setEncoding(XML_ENCODING);
		config.setAutoSave(true);
		
		config.addNodes(null, nodes);
		
		writeConfigurationFormatted(config, file);
	}
	
	
	protected void writeConfigurationFormatted(XMLConfiguration xml, File file) {
		
		try
		{
			// Save unformatted data to dummy-out, which creates our document as side-effect...
			ByteArrayOutputStream bOut = new ByteArrayOutputStream();
			xml.save(new OutputStreamWriter(bOut, XML_ENCODING));
			
			Document doc = xml.getDocument();
			
			
			// Format document
			DOMImplementationLS domLS = (DOMImplementationLS) doc.getImplementation();

			LSSerializer serializer = domLS.createLSSerializer();
			
			DOMConfiguration domConfig = serializer.getDomConfig();
			if (domConfig.canSetParameter("format-pretty-print", Boolean.TRUE))
			{
				domConfig.setParameter("format-pretty-print", Boolean.TRUE);
			}
			serializer.setNewLine("\n");
			
			
			// Prepare output...
			FileOutputStream fOut = new FileOutputStream(file);

			LSOutput output = domLS.createLSOutput();
			output.setEncoding(XML_ENCODING);
			output.setByteStream(fOut);
		   
			
			// ...and finally write!
			serializer.write(doc, output);

			fOut.close();
		} catch (IOException err)
		{
			log.error("Error while saving file!", err);
		} catch (ConfigurationException err)
		{
			log.error("Error while buffering (unformatted) config!", err);
		}
	}
}
