package edu.dhbw.mannheim.tigers.sim.engine.plugins;


import java.io.File;
import java.io.FilenameFilter;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.SubnodeConfiguration;
import org.apache.commons.configuration.XMLConfiguration;
import org.apache.commons.configuration.tree.ConfigurationNode;
import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.Sim;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.cam.ACamOutputPlugin;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.cam.DefaultTwoCamOutput;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.visualization.AVisualizationPlugin;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.visualization.DefaultVisualizationPlugin;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimPluginException;


/**
 * This class loads and grants access to the plugins which are lying in {@link Cage#PLUGIN_PATH}.
 * 
 * @author Gero
 */
public class PluginProvider implements IPluginProvider {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// Constants
	private static final String								PATH							= Sim.PLUGIN_PATH;
	private static final String								DEFAULT_PROPERTIES_PATH	= Sim.PROPERTIES_PATH
																												+ "defaultplugins.xml";
	private static final String								PLUGIN_CONFIG				= "plugin.xml";
	private static final String								EMPTY_FQN					= "EMPTY_FQN";
	
	private static PluginProvider								instance						= null;
	
	private final Logger											log							= Logger.getLogger(getClass());
	

	private final PluginConfigHelper							configHelper				= new PluginConfigHelper();
	
	// Content
	private final Map<PluginID, ClassLoader>				classLoaders				= new HashMap<PluginID, ClassLoader>();
	
	private final Map<PluginID, AVisualizationPlugin>	visPlugins					= new HashMap<PluginID, AVisualizationPlugin>();
	private final Map<PluginID, ACamOutputPlugin>		camPlugins					= new HashMap<PluginID, ACamOutputPlugin>();
	// private final Map<PluginID, ASimPlugin> controlPlugins = new HashMap<PluginID, ASimPlugin>();
	
	private final Map<PluginID, String>						configPaths					= new HashMap<PluginID, String>();
	
	private final Map<EPluginType, PluginID>				defaults						= new HashMap<EPluginType, PluginID>();
	private File													defaultsConfigFile		= null;
	

	// Support
	// private final ClazzFileFilter clazzFilter = new ClazzFileFilter();
	private final JarZipFileFilter							jarOrZipFilter				= new JarZipFileFilter();
	private final NoHiddenFolderFilter						noHiddenFolderFilter		= new NoHiddenFolderFilter();
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	  * @see PluginProvider
	  */
	private PluginProvider() {
	}
	

	static synchronized PluginProvider getInstance() {
		if (instance == null) {
			instance = new PluginProvider();
		}
		return instance;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public synchronized void loadPlugins() {
		
		File pluginFolder = new File(PATH);
		if (!checkFolder(pluginFolder)) {
			log.debug("Can't access '" + pluginFolder.getAbsolutePath() + "'; no permission??");
			return;
		}
		
		File[] plugins = pluginFolder.listFiles(noHiddenFolderFilter);
		if (plugins.length == 0) {
			log.debug("No plugins found!");
			return;
		}
		
		// Dig through the different plugin-folders
		for (File onePluginFolder : plugins) {
			
			if (checkFolder(onePluginFolder)) {
				try {
					// ### Load manifest-file
					File configFile = new File(onePluginFolder.getPath() + "/" + PLUGIN_CONFIG);
					checkFileAvailable(configFile);
					XMLConfiguration xml = new XMLConfiguration(configFile);
					
					List<?> pluginNodes = xml.configurationsAt("plugin");
					
					
					// ### Create URLClassLoader
					List<URL> urls = new ArrayList<URL>();
					
					// Add plugin-folder (for .class-files)
					// URLClassLoader treats URLs without trailing '/' as Jars!
					String path = onePluginFolder.getAbsolutePath() + "/";
					urls.add(new URL("file:///" + path));
					
					// Add any .jar-files in the plugin-folder
					File[] libFiles = onePluginFolder.listFiles(jarOrZipFilter);
					for (File lib : libFiles) {
						if (!checkFileAvailable(lib)) {
							log.warn("Jar/Zip '" + lib.getName() + "' can not be accessed!");
						} else {
							urls.add(new URL("file:///" + lib.getAbsolutePath()));
						}
					}
					
					// Create!
					URL[] urlArr = urls.toArray(new URL[urls.size()]);
					URLClassLoader ucl = new URLClassLoader(urlArr, getClass().getClassLoader());
					
					
					// ### Iterate over declared plugins
					for (Object obj : pluginNodes) {
						SubnodeConfiguration pluginNode = (SubnodeConfiguration) obj;
						String fqn = pluginNode.getString("fqn", EMPTY_FQN);
						
	
						// ### Load plugin-class
						loadPlugin(ucl, fqn, pluginNode, configFile.getPath());
					}
					
				} catch (MalformedURLException mue) {
					log.error("Unproper URL: " + mue.getMessage());
					continue;
				} catch (ConfigurationException err) {
					log.error("Error while reading plugin-configuration: ", err);
					continue;
				}
			} else {
				log.warn("Folder '" + onePluginFolder.getName()
						+ "' can not be loaded; no permission?? (This will be skipped)");
				continue;
			}
		}
		

		// Finally, add default-plugins
		loadDefaultPlugins();
	}
	

	/**
	 * Registers default plugins for every type to allow fail-safe operation!
	 */
	private void loadDefaultPlugins() {
		
		// Read from properties file
		defaultsConfigFile = new File(DEFAULT_PROPERTIES_PATH);
		XMLConfiguration defaultsConfig = configHelper.createAndcheckDefaultsProperties(defaultsConfigFile);
		

		this.defaults.put(EPluginType.CAM_OUTPUT, DefaultTwoCamOutput.ID);
		SubnodeConfiguration camConfig = configHelper.getPluginConfigById(defaultsConfig, DefaultTwoCamOutput.ID);
		DefaultTwoCamOutput defCamOut = new DefaultTwoCamOutput(camConfig);
		this.camPlugins.put(defCamOut.getID(), defCamOut);
		log.info("Successfully loaded (default) " + defCamOut + ".");
		
		this.defaults.put(EPluginType.VISUALIZATION, DefaultVisualizationPlugin.ID);
		SubnodeConfiguration visConfig = configHelper.getPluginConfigById(defaultsConfig, DefaultVisualizationPlugin.ID);
		DefaultVisualizationPlugin defVis = new DefaultVisualizationPlugin(visConfig);
		this.visPlugins.put(defVis.getID(), defVis);
		log.info("Successfully loaded (default) " + defVis + ".");
	}
	

	/**
	 * Loads the class specified by fqn using the given class-loader, and introduces it to the
	 * context of the {@link PluginProvider}
	 * 
	 * @param cll ClassLoader-instance
	 * @param fqn Fully qualified name
	 */
	private void loadPlugin(ClassLoader cll, String fqn, SubnodeConfiguration config, String configPath) {
		
		try {
			Class<?> clazz = cll.loadClass(fqn);
			
			if (clazz == null) {
				log.debug("No class found for fqn='" + fqn + "'!");
				return;
			}
			
			if (!ASimPlugin.class.isAssignableFrom(clazz)) {
				log.error("'" + clazz + "' is not a subclass of 'ASimPlugin'!");
				return;
			}
			
			// Instantiate
			Constructor<?> constructor = clazz.getConstructor(SubnodeConfiguration.class);
			ASimPlugin plugin = (ASimPlugin) constructor.newInstance(config);
			

			// Handle plugin
			Map<PluginID, ASimPlugin> map = getPluginMap(plugin.getType());
			
			PluginID id = plugin.getID();
			ASimPlugin firstPlugin = map.get(id);
			if (firstPlugin != null) {
				log.warn("A plugin with id " + plugin.getID() + " (" + firstPlugin
						+ ") has already been loaded; this will be skipped!");
				return;
			} else {
				map.put(id, plugin);
				configPaths.put(id, configPath);
			}
			
			classLoaders.put(id, cll); // Store for reuse (reloading etc.)
			
			log.info("Successfully loaded " + plugin + ".");
			

		} catch (ClassNotFoundException cnf) {
			log.error("Plugin-class could not be found! (" + fqn + ")");
		} catch (SecurityException se) {
			log.error("The chosen constructor may not be accessed: " + fqn, se);
		} catch (NoSuchMethodException nsm) {
			log.error("The chosen constructor does not exist: " + fqn, nsm);
		} catch (IllegalArgumentException iae) {
			log.error("Error while creating plugin-instance: " + fqn, iae);
		} catch (InstantiationException ine) {
			log.error("Error while creating plugin-instance: " + fqn, ine);
		} catch (IllegalAccessException iae) {
			log.error("Error while creating plugin-instance: " + fqn, iae);
		} catch (InvocationTargetException ite) {
			log.error("Error while creating plugin-instance: " + fqn, ite);
		} catch (SimPluginException err) {
			log.error("Error while handling created plugin: " + fqn, err);
		} catch (AbstractMethodError ame) {
			log.error("Error while calling plugin-constructor: " + fqn, ame);
		}
	}
	

	/**
	 * @param <T> The type which extends {@link ASimPlugin}
	 * @param type {@link EPluginType}
	 * @return The proper map for the given {@link EPluginType}.
	 * @throws SimPluginException
	 */
	@SuppressWarnings({ "unchecked" })
	private synchronized <T extends ASimPlugin> Map<PluginID, T> getPluginMap(EPluginType type)
			throws SimPluginException {
		
		switch (type) {
			case CAM_OUTPUT:
				return (Map<PluginID, T>) camPlugins;
				
			case VISUALIZATION:
				return (Map<PluginID, T>) visPlugins;
				
			default:
				throw new SimPluginException("Unnable to handle plugin-type: '" + type + "'!");
		}
	}
	

	/**
	 * @param file
	 * @return True if the folder exists and user has r/w-access
	 */
	private boolean checkFileAvailable(File file) {
		if (!file.exists() || !file.canRead() || !file.canWrite() || file.isDirectory()) {
			return false;
		}
		return true;
	}
	

	/**
	 * @param folder
	 * @return True if the folder exists, user has r/w-access and is a folder
	 */
	private boolean checkFolder(File folder) {
		if (!folder.exists() || !folder.canRead() || !folder.canWrite() || !folder.isDirectory()) {
			return false;
		}
		return true;
	}
	

	@Override
	public void saveLoadedPluginsProperties() {
		savePluginsProperties(false);
	}
	

	/**
	 * Saves the properties of every loaded plugin
	 * 
	 * @param removeDefaults Whether the default-plugins should be removed from the plugin-maps (used
	 *           for deinitialization/unloading)
	 */
	private void savePluginsProperties(boolean removeDefaults) {
		
		// Handle default plugin (especially their configuration)
		List<ConfigurationNode> nodes = new ArrayList<ConfigurationNode>();
		
		// Gather configurations from default-plugins
		for (Entry<EPluginType, PluginID> entry : defaults.entrySet()) {
			
			// Get config
			ASimPlugin plugin = getPlugin(entry.getValue(), entry.getKey());
			nodes.add(plugin.getConfig().getRootNode());
			
			if (removeDefaults) {
				// Remove default-plugins from maps, as their config is saved now...
				try {
					Map<PluginID, ASimPlugin> map = getPluginMap(plugin.getType());
					map.remove(plugin.getID());
				} catch (SimPluginException err) {
					// This is veeery unlikely...
				}
			}
		}
		
		try {
			configHelper.savePluginNodesToFile(nodes, defaultsConfigFile);
		} catch (ConfigurationException err) {
			log.error("Unable to write defaults-config: ", err);
		}
		

		// Iterate over all plugins which are not default
		for (int i = 0; i < EPluginType.values().length; i++) {
			// Dynamically clear all maps associated with a EPluginType
			EPluginType type = EPluginType.values()[i];
			
			try {
				Map<PluginID, ASimPlugin> map = getPluginMap(type);
				
				// Save their configuration
				for (ASimPlugin plugin : map.values()) {
					
					// Check whether this is a default-plugin
					PluginID defID = defaults.get(type);
					if (defID != null && defID.equals(plugin.getID())) { // Should be none-null but...
						continue; // Default, has already been handled above! =)
					}
					
					// Retrieve config-path..
					String path = configPaths.get(plugin.getID());
					if (path == null) {
						log.debug("The path of a plugin is 'null', this usually means that someone added for then one default-plugin per EPluginType! ("
								+ plugin + ") This plugins settings will not be saved!");
						continue;
					}
					File configFile = new File(path);
					checkFileAvailable(configFile);
					
					List<ConfigurationNode> node = new ArrayList<ConfigurationNode>();
					node.add(plugin.getConfig().getRootNode());
					
					// ... and store it there
					try {
						configHelper.savePluginNodesToFile(node, configFile);
					} catch (ConfigurationException err) {
						log.error("Unable to write config: ", err);
					}
				}
			} catch (SimPluginException err) {
				// If the given type is not supported (yet)
			}
		}
	}
	

	private void unloadPlugins() {
		
		savePluginsProperties(true);
		

		// Finally, clear all maps
		for (int i = 0; i < EPluginType.values().length; i++) {
			// Dynamically clear all maps associated with a EPluginType
			EPluginType type = EPluginType.values()[i];
			
			try {
				Map<PluginID, ASimPlugin> map = getPluginMap(type);
				map.clear();
			} catch (SimPluginException err) {
				// If the given type is not supported (yet)
			}
		}
		

		// Clear content
		classLoaders.clear();
		defaults.clear();
		configPaths.clear();
		
		defaultsConfigFile = null;
		

		log.info("Plugins successfully unloaded!");
	}
	

	@Override
	public synchronized void reloadPlugins() {
		
		unloadPlugins();
		
		// ========== After this line, default state should be reestablished
		
		loadPlugins();
	}
	

	@Override
	public void cleanup() {
		unloadPlugins();
	}
	

	@Override
	public synchronized List<PluginID> getLoadedPluginsOf(EPluginType type) throws SimPluginException {
		Map<PluginID, ASimPlugin> map = getPluginMap(type);
		return new ArrayList<PluginID>(map.keySet());
	}
	

	@Override
	public List<ASimPlugin> getLoadedPlugins() {
		List<ASimPlugin> result = new ArrayList<ASimPlugin>();
		
		for (int i = 0; i < EPluginType.values().length; i++) {
			EPluginType type = EPluginType.values()[i];
			
			try {
				Map<PluginID, ASimPlugin> map = getPluginMap(type);
				result.addAll(map.values());
			} catch (SimPluginException err) {
				log.error(err.getMessage(), err);
			}
		}
		
		return result;
	}
	

	@SuppressWarnings("unchecked")
	@Override
	public synchronized <T extends ASimPlugin> T getPlugin(PluginID id, EPluginType type) {
		
		ASimPlugin result = null;
		try {
			Map<PluginID, ASimPlugin> map = getPluginMap(type);
			
			if (id != null) {
				result = map.get(id);
			}
			
			if (result == null) {
				PluginID defaultID = defaults.get(type);
				if (defaultID == null) {
					log.debug("Default for type '" + type + "' not set!!!");
				}
				
				result = map.get(defaultID);
				if (result == null) {
					log.debug("Default plugin with id " + defaultID + " not available!");
				}
				log.warn("Plugin of type '" + type + "' and with id " + id + " is not available, took " + result
						+ " instead!");
			}
		} catch (SimPluginException spe) {
			// type cannot be handled, but is specified in EPluginType;
			// or default
		}
		
		return (T) result;
	}
	
	
	private class JarZipFileFilter implements FilenameFilter {
		
		@Override
		public boolean accept(File dir, String name) {
			return name.toLowerCase().endsWith(".jar") || name.toLowerCase().endsWith(".zip");
		}
	}
	
	
	private class NoHiddenFolderFilter implements FilenameFilter {
		
		@Override
		public boolean accept(File dir, String name) {
			return !dir.isHidden() && !name.toLowerCase().startsWith(".");
		}
	}
	

	// private class ClazzFileFilter implements FilenameFilter {
	//
	// @Override
	// public boolean accept(File dir, String name) {
	// return name.toLowerCase().endsWith(".class");
	// }
	// }
}
