/* 
 * *********************************************************
 * Copyright (c) 2009 - 2011, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: 24.05.2011
 * Author(s): Gero
 *
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sim.engine.plugins.cam;

import com.jme.math.FastMath;

import edu.dhbw.mannheim.tigers.sim.engine.model.CoordinateConversionHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.TimedModelLocation;
import edu.dhbw.mannheim.tigers.sumatra.model.data.MessagesRobocupSslDetection.SSL_DetectionBall;
import edu.dhbw.mannheim.tigers.sumatra.model.data.MessagesRobocupSslDetection.SSL_DetectionRobot;

/**
 * This class is the basic representation of a camera as it is necessary for the assembly of SSL-Vision frames.
 * 
 * @author Gero
 */
public abstract class ACam {
	
	private final int id;
	
	private long packetCount = 0;
	
	
	/**
	 * @param id
	 */
	public ACam(int id) {
		this.id = id;
	}
	
	
	/**
	 * Increases the packet-count by one
	 * 
	 * @return The OLD packet-count
	 */
	public long incPacketCount() {
		return ++packetCount;
	}
	
	
	/**
	 * @param id
	 * @param obj
	 * @return A {@link SSL_DetectionRobot}-representation of a simulated robot. <code>Null</code> if the camera does
	 * not see this robot!!!
	 */
	public SSL_DetectionRobot.Builder createRobotDetection(int id, TimedModelLocation bot) {
		if (!isObjVisible(bot)) {
			return null;
		}
		
		return convertRobot(id, bot);
	}
	
	
	/**
	 * @param id
	 * @param obj
	 * @return A {@link SSL_DetectionRobot}-representation of a simulated robot.
	 */
	protected SSL_DetectionRobot.Builder convertRobot(int id, TimedModelLocation bot) {
		SSL_DetectionRobot.Builder b = SSL_DetectionRobot.newBuilder();
		
		// Which bot?
		b.setRobotId(id)

		// Info-quality
		.setConfidence(1.0f)
		// Absolute
		.setHeight(0)
		// WTF???
		.setPixelX(0)
		// WTF???
		.setPixelY(0)
		// WTF???
		
		// Position
		.setX(bot.getZmm()).setY(bot.getXmm())

		// Orientation
		.setOrientation(convertOrientation(bot.getOrientation()));
		
		return b;
	}
	
	
	/**
	 * @param obj
	 * @return A {@link SSL_DetectionBall}-representation of a simulated ball. <code>Null</code> if the camera does
	 * not see this ball!!!
	 */
	public SSL_DetectionBall.Builder createBallDetection(TimedModelLocation ball) {
		if (!isObjVisible(ball)) {
			return null;
		}
		
		return convertBall(ball);
	}
	
	
	/**
	 * @param obj
	 * @return A {@link SSL_DetectionBall}-representation of a simulated ball.
	 */
	protected SSL_DetectionBall.Builder convertBall(TimedModelLocation ball) {
		SSL_DetectionBall.Builder b = SSL_DetectionBall.newBuilder();
		
		b.setArea(0) // WTF?
		.setConfidence(1.0f) // Absolute
		.setPixelX(0) // WTF?
		.setPixelY(0) // WTF?
		
		.setX(ball.getZmm()).setY(ball.getXmm())
		//TODO remove this workaround when implementing real cameras providing 
		//     correct z-position data.
		//     When using ball.getYmm(), this outputs the center of the ball, 
		//     which is not on the field when the ball is rolling but 21.466162 
		//     above.
		//.setZ(ball.getYmm()
		.setZ(0.0f);
		return b;
	}
	
	
	/**
	 * @param simOrientation
	 * @return The orientation turned to SSL-Vision orientation.
	 */
	protected float convertOrientation(float simOrientation) {
		return CoordinateConversionHelper.shiftRadians(simOrientation, FastMath.HALF_PI);
	}

	
	/**
	 * @param obj
	 * @return Whether this obj is visible to this {@link ACam}
	 */
	protected abstract boolean isObjVisible(TimedModelLocation obj);
	
	
	/**
	 * @return the id
	 */
	public int getId() {
		return this.id;
	}
	
	
	/**
	 * @param packetCount the packetCount to set
	 */
	public void setPacketCount(long packetCount) {
		this.packetCount = packetCount;
	}
	
	
	/**
	 * @return the packetCount
	 */
	public long getPacketCount() {
		return this.packetCount;
	}
}
