package edu.dhbw.mannheim.tigers.sim.settings;


import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.ExtendedMetaData;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.emf.ecore.xmi.impl.XMLResourceImpl;

import edu.dhbw.mannheim.tigers.sim.Sim;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimSettingsException;
import edu.dhbw.mannheim.tigers.sim.resources.AResourceProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.IApplicationProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ICamProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.IDataFactory;
import edu.dhbw.mannheim.tigers.sim.settings.data.IDataPackage;
import edu.dhbw.mannheim.tigers.sim.settings.data.IMatchModeProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.IShooterModeProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ISimEngineProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ISimProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ITeamProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ITestModeProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.IVisualizationProperties;


/**
 * This class provides application-wide access to the application- and simulation-settings
 * 
 * @author Gero
 */
public class SettingsProvider extends AResourceProvider<String, ISimProperties> implements ISettingsProvider {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static SettingsProvider				instance			= null;
	
	private final IDataFactory						factory			= IDataFactory.eINSTANCE;
	
	// Store
	private ISimProperties							properties;
	
	// Persistence
	private static final Map<String, Object>	WRITE_OPTIONS	= new HashMap<String, Object>();
	
	static {
		WRITE_OPTIONS.put(XMLResource.OPTION_KEEP_DEFAULT_CONTENT, Boolean.TRUE); // Write default data to file
		WRITE_OPTIONS.put(XMLResource.OPTION_SAVE_ONLY_IF_CHANGED, Boolean.TRUE);
		WRITE_OPTIONS.put(XMLResource.OPTION_FORMATTED, Boolean.TRUE);
		WRITE_OPTIONS.put(XMLResource.OPTION_LINE_WIDTH, 1);
	}
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see SettingsProvider
	 */
	public SettingsProvider() {
		super(SettingsProvider.class);
	}
	

	public static synchronized SettingsProvider getInstance() {
		if (instance == null) {
			instance = new SettingsProvider();
		}
		return instance;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void doInitialize() {
		
		@SuppressWarnings("unused")
		IDataPackage pack = IDataPackage.eINSTANCE; // Register package (done as side effect), as this
																	// is no RCP environment
	}
	

	@Override
	protected void doLoad() {
		
		// Load settings from file using EMF
		Resource resource = new XMLResourceImpl();
		
		Map<String, ExtendedMetaData> options = new HashMap<String, ExtendedMetaData>();
		options.put(XMLResource.OPTION_EXTENDED_META_DATA, ExtendedMetaData.INSTANCE);
		

		File settingsFile = new File(Sim.SETTINGS_PATH);
		
		try {
			resource.load(new FileInputStream(settingsFile), options);
			EObject result = resource.getContents().get(0);
			
			if (result instanceof ISimProperties) {
				properties = (ISimProperties) result;
			} else {
				log.warn("Could not load the ISimProperties, generating a default set of settings!");
			}
		} catch (FileNotFoundException exc) {
			log.warn("Couldn't locate settings file '" + settingsFile.getAbsolutePath()
					+ "'; generating new, default settings!");
			
		} catch (IOException exc) {
			log.warn("Couldn't load settings from file '" + settingsFile.getAbsolutePath()
					+ "'; generating new, default settings!");
		} finally {
			checkSettingsConsistency();
		}
	}
	

	/**
	 * Checks whether {@link #properties} contains a proper set of properties. If something is missing (e.g., after an
	 * extension of the properties), a default properties-branch for this part is created and added. After calling this
	 * method the properties are guaranteed to be complete!
	 */
	private void checkSettingsConsistency() {
		
		if (properties == null) {
			properties = factory.createSimProperties();
		}
		
		ICamProperties camProps = properties.getCamProperties();
		if (camProps == null) {
			camProps = factory.createCamProperties();
			properties.setCamProperties(camProps);
		}
		
		ISimEngineProperties engineProps = properties.getSimEngineProperties();
		if (engineProps == null) {
			engineProps = factory.createSimEngineProperties();
			properties.setSimEngineProperties(engineProps);
		}
		
		IApplicationProperties appProps = properties.getApplicationProperties();
		if (appProps == null) {
			appProps = factory.createApplicationProperties();
			properties.setApplicationProperties(appProps);
		}
		
		IVisualizationProperties visProps = properties.getVisualizationProperties();
		if (visProps == null) {
			visProps = factory.createVisualizationProperties();
			properties.setVisualizationProperties(visProps);
		}
		
		// Mode properties
		IMatchModeProperties robocupMode = properties.getMatchModeProperties();
		if (robocupMode == null) {
			robocupMode = factory.createMatchModeProperties();
			properties.setMatchModeProperties(robocupMode);
		}
		
		// Teams
		ITeamProperties teamYellow = robocupMode.getTeamYellow();
		if (teamYellow == null) {
			teamYellow = factory.createTeamProperties();
			teamYellow.setBotIDs("1,2,3,4,5");
			teamYellow.setFirstBotInputPort(10000);
			teamYellow.setMcastDiscoverGroup("225.42.42.42");
			teamYellow.setMcastListenPort(10041);
			teamYellow.setMcastServerPort(10040);
			robocupMode.setTeamYellow(teamYellow);
		}
		
		ITeamProperties teamBlue = robocupMode.getTeamBlue();
		if (teamBlue == null) {
			teamBlue = factory.createTeamProperties();
			teamBlue.setBotIDs("6,7,8,9,10");
			teamBlue.setFirstBotInputPort(10100);
			teamBlue.setMcastDiscoverGroup("225.42.42.42");
			teamBlue.setMcastListenPort(10141);
			teamBlue.setMcastServerPort(10140);
			robocupMode.setTeamBlue(teamBlue);
		}
		
		IShooterModeProperties shooterMode = properties.getShooterModeProperties();
		if (shooterMode == null) {
			shooterMode = factory.createShooterModeProperties();
			properties.setShooterModeProperties(shooterMode);
		}
		
		ITestModeProperties testMode = properties.getTestModeProperties();
		if (testMode == null) {
			testMode = factory.createTestModeProperties();
			properties.setTestModeProperties(testMode);
		}
	}
	

	@Override
	public ISimProperties getSimProperties() throws SimSettingsException {
		
		synchronized (storeLock) {
			if (properties == null) {
				throw new SimSettingsException("No settings available!");
			}
			
			return properties;
		}
	}
	

	@Override
	public void persist() {
		synchronized (storeLock) {
			save();
		}
	}
	

	@Override
	public void cleanup() {
		
		synchronized (storeLock) {
			
			save();
			
			super.cleanup();
			
			properties = null;
		}
	}
	

	private void save() {
		
		synchronized (storeLock) {
			
			// Save ISimProperties using EMF
			Resource resource = new XMLResourceImpl();
			resource.getContents().add(properties);
			
			File file = new File(Sim.SETTINGS_PATH);
			
			try {
				
				if (!file.exists()) {
					if (!file.createNewFile()) {
						log.error("Couldn't create settings-file to '" + file.getAbsolutePath() + "'; they get lost...");
					}
				}
				
				FileOutputStream fos = new FileOutputStream(file);
				resource.save(fos, WRITE_OPTIONS);
				fos.flush();
				fos.close();
				log.info("Successfully saved application settings!");
				
			} catch (FileNotFoundException exc) {
				log.error("Couldn't write settings to '" + file.getAbsolutePath() + "'...", exc);
			} catch (IOException exc) {
				log.error("Couldn't write settings to '" + file.getAbsolutePath() + "'...", exc);
			}
		}
	}
}
