package edu.dhbw.mannheim.tigers.sim.util.activeobserver;


import java.util.LinkedList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.log4j.Logger;


/**
 * <p>
 * An implementation of the Observer-Pattern.
 * </p>
 * <p>
 * The idea of this implementation is to not consume the Observable's Thread by letting a new Thread handle and process
 * all calls of the Observable's {@link #update(IActiveObservable, Object)} -method, or even start a new Thread for each
 * call.
 * </p>
 * <p>
 * On a call of the {@link #update(IActiveObservable, Object)}-method, the parameters are stored in the
 * {@link #eventPipe}; and the {@link #eventHandler} is notified that there is a new call. It fetches the call, and
 * works it by itself - or let a {@link CallWorker} do the job; depends on the {@link #oneThreadPerCall}-switch.
 * </p>
 * 
 * @param <S> Observable type (Subject)
 * @param <O> Observer type
 * @param <E> Event type
 * 
 * @author Gero
 */
public abstract class AActiveObserver<S extends IActiveObservable<S, O, E>, O extends IActiveObserver<S, O, E>, E>
		implements IActiveObserver<S, O, E> {
	
	private final Logger									log					= Logger.getLogger("ActiveObservers_EventHandler");
	
	private String											name					= "";
	
	// Event pipe
	private final LinkedList<UpdateCall<S, E>>	eventPipe			= new LinkedList<UpdateCall<S, E>>();
	

	// Initialization
	private boolean										alreadyStarted		= false;
	private final Object									startedSync			= new Object();
	private int												observableCount	= 0;
	

	// Synchronization and event handling
	private final Object									eventSync			= new Object();
	private final EventHandler							eventHandler		= new EventHandler();
	private boolean										oneThreadPerCall	= false;
	

	// Event
	private ExecutorService								exService			= null;
	

	// Flags
	/** Whether incoming events should be blocked or not */
	private boolean										blockedFlag			= false;
	
	
	private synchronized ExecutorService exService() {
		if (exService == null) {
			exService = Executors.newCachedThreadPool();
		}
		return exService;
	}
	

	private void execute(Runnable runnable) {
		exService().execute(runnable);
	}
	

	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see AActiveObserver#AActiveObserver(ExecutorService, boolean)
	 */
	public AActiveObserver(boolean oneThreadPerCall, String name) {
		this(null, oneThreadPerCall, name);
	}
	

	/**
	 * Watch oneThreadPerCall! If it's <code>true</code>, the code you put into
	 * {@link AActiveObserver#onUpdate(IActiveObservable, Object)} must be thread-safe!!!
	 * 
	 * @param newExService The {@link ExecutorService} the Observer should use. If <code>null</code>,
	 *           it takes its own.
	 * @param oneThreadPerCall Very important
	 * @param name The name of the EventHandler-Thread
	 */
	public AActiveObserver(ExecutorService newExService, boolean oneThreadPerCall, String name) {
		this.exService = newExService;
		
		this.oneThreadPerCall = oneThreadPerCall;
		this.name = name;
	}
	
	
	// --------------------------------------------------------------------------
	// --- event handling -------------------------------------------------------
	// --------------------------------------------------------------------------
	private class EventHandler implements Runnable {
		
		/** Flag that indicates that the EventHandler should shut down */
		private boolean	shutdown				= false;
		/** Flag that */
		private boolean	shutdownOnEmpty	= false;
		
		
		@Override
		public void run() {
			
			Thread.currentThread().setName(name);
			
			UpdateCall<S, E> currentCall = null;
			
			try {
				while (!shutdown) {
					
					synchronized (eventSync) {
						
						if (eventPipe.isEmpty()) {
							if (shutdownOnEmpty) {
								return;
							} else {
								eventSync.wait();
								continue; // New loop to force checks!
							}
						}
						
						// Fetch call from stack
						currentCall = eventPipe.pollFirst();
					}
					
					if (oneThreadPerCall) {
						execute(new CallWorker<S, E>(currentCall.observable, currentCall.event));
					} else {
						onUpdate(currentCall.observable, currentCall.event);
					}
					
				}
			} catch (InterruptedException exc) {
				if (!shutdown) {
					log.error("Error while waiting for new incoming event!", exc);
				}
			} finally {
				synchronized (startedSync) {
					alreadyStarted = false;
				}
				
				if (shutdown || shutdownOnEmpty) {
					log.debug(name + " shutting down...");
				}
			}
		}
		

		public void shutdown() {
			this.shutdown = true;
		}
		

		public void shutdownOnEmpty() {
			this.shutdownOnEmpty = true;
		}
	}
	
	
	/**
	 * <b>Attention:</b> If oneThreadPerCall is <code>true</code>, this code must be
	 * <b>thread-safe!</b>
	 * 
	 * @param observable
	 * @param event
	 */
	protected abstract void onUpdate(S observable, E event);
	

	// --------------------------------------------------------------------------
	// --- life cycle -----------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void update(S observable, E event) {
		
		if (!blockedFlag) {
			
			synchronized (eventSync) {
				eventPipe.add(new UpdateCall<S, E>(observable, event));
				
				eventSync.notifyAll();
			}
		}
	}
	

	@Override
	public void onAdd() {
		
		synchronized (startedSync) {
			
			observableCount++;
			
			if (!alreadyStarted) {
				execute(eventHandler);
				alreadyStarted = true;
			}
		}
	}
	

	@Override
	public void onRemove() {
		
		synchronized (startedSync) {
			if (!alreadyStarted) {
				return; // SHOULD not occur...
			}
			
			observableCount--;
		}
	}
	

	/**
	 * Calls {@link AActiveObserver#shutdown(boolean, boolean, boolean)} with <code>false, true, true</code>
	 */
	public void shutdown() {
		shutdown(false, true, true);
	}
	

	/**
	 * Closes the {@link AActiveObserver} and closes all threads
	 * 
	 * @param forceShutdown Forces a hard shutdown of the observer, canceling all current events and
	 *           interrupting the {@link EventHandler} and all CallWorkers (if there are any)
	 * @param finishCurrentEvents If <code>forceShutdown == false</code>, this indicates whether to
	 *           process all events until the pipe is empty, or to skip after the one
	 *           that is currently processed
	 * @param blockNewEvents Whether new events are enqueued or not (If <code>forceShutdown == false</code> and
	 *           <code>finishCurrentEvents == true</code>,
	 *           possibility to let the Observer run dry and die!)
	 */
	public void shutdown(boolean forceShutdown, boolean finishCurrentEvents, boolean blockNewEvents) {
		
		synchronized (startedSync) {
			
			if (forceShutdown) {
				eventHandler.shutdown(); // Set shutdown-flag
				exService().shutdown(); // Interrupt everything
			} else { // or...
				eventHandler.shutdownOnEmpty(); // Shutdown only on empty list
				if (!finishCurrentEvents) {
					eventHandler.shutdown(); // AND after the current event has been processed
				} else {
					if (blockNewEvents) {
						blockedFlag = true; // Block incoming events
					}
				}
			}
			
			synchronized (eventSync) {
				eventSync.notifyAll(); // Wake up, in case EventHandler slept while has been murdered...
			}
		}
	}
	
	
	// --------------------------------------------------------------------------
	// --- inner classes --------------------------------------------------------
	// --------------------------------------------------------------------------
	private class UpdateCall<SC extends IActiveObservable<S, O, E>, EC> {
		
		public final SC	observable;
		public final EC	event;
		
		
		public UpdateCall(SC newObservable, EC event) {
			this.observable = newObservable;
			this.event = event;
		}
	}
	
	
	private class CallWorker<SC extends S, EC extends E> implements Runnable {
		
		private final SC	observable;
		private final EC	event;
		
		
		public CallWorker(SC observable, EC event) {
			this.observable = observable;
			this.event = event;
		}
		

		@Override
		public void run() {
			
			Thread.currentThread().setName("Call_Worker");
			
			onUpdate(observable, event);
		}
	}
}
