/**
 * 
 */
package edu.dhbw.mannheim.tigers.sim.util.network;


import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.net.DatagramPacket;
import java.net.InetAddress;

import org.apache.log4j.Logger;


/**
 * This class should be able to join a multicast group, catch packets and forward them if they are
 * of the given type.
 * 
 * @author Gero
 * 
 */
public class ObjectReceiver<M extends Serializable, I extends IObjectMessageInterpreter<M>> {

	// Util
	private Logger log = Logger.getLogger(getClass());

	// Communication
	private static final int STAN_BUFFER_SIZE = 1024;
	private int bufferSize = STAN_BUFFER_SIZE;
	private BufferedInputStream byteIn = null;
	private ObjectInputStream inStream = null;

	// Receiver
	private IReceiver receiver;
	private I interpreter = null;
	
	
	private boolean shutdown = false;
	
	
	public ObjectReceiver(I interpreter, IReceiver receiver) {
		this(STAN_BUFFER_SIZE, interpreter, receiver);
	}

	public ObjectReceiver(int bufferSize, I interpreter, IReceiver receiver) {
		
		this.bufferSize = bufferSize;
		this.interpreter = interpreter;
		this.receiver = receiver;
		
	}
	
	
	@SuppressWarnings("unchecked")
	public void listen() {
		
		log.debug("ObjectReceiver starts listening.");
		
		byte[] buffer = null;
		DatagramPacket packet = null;
		byte[] data = null;
		
		while (!Thread.currentThread().isInterrupted() && !shutdown) {
			buffer = new byte[bufferSize];
			packet = new DatagramPacket(buffer, buffer.length);
			
			// Receive packet
			try {
				packet = receiver.receive(packet);
				
				if (packet == null) {
					shutdown = true;
					continue;
				}
			} catch (IOException err) {
				if (shutdown) {
					break;
				} else {
					log.error("Error while receiving new package", err);
					continue;
				}
			}
			
			// Byte to In-Stream
			data = packet.getData();
			byteIn = new BufferedInputStream(new ByteArrayInputStream(data, packet.getOffset(), packet.getLength()));
			try {
				inStream = new ObjectInputStream(byteIn);
			} catch (IOException err1) {
				log.error("Error while creating an ObjectInputStream!", err1);
				continue;
			}
			
			// In-Stream to Object
			Object obj = null;
			try {
				obj = inStream.readObject();
				inStream.close();
			} catch (IOException err) {
				log.error("Error while reading Object from InputStream; maybe the buffer is to small??", err);
				continue;
			} catch (ClassNotFoundException err) {
				log.warn("Class could not be found!", err);
			}
			
			// Object to M
			M msg = null;
			try {
				msg = (M) obj;
			} catch (ClassCastException err) {
				log.warn("Could not cast '" + obj + "'.", err);
			}
			
			if (msg != null) {
				interpreter.receive(msg);
			}
		}
		
		
		log.debug("ObjectReceiver stopped listening.");
	}
	
	
	public void shutdown() {
		shutdown = true;
		
		if (receiver != null) {
			try {
            receiver.cleanup();
         } catch (IOException err) {
            log.debug("Interrupted while receiving...");
         }
		}
	}

	
	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	public int getPort() {
		return this.receiver.getLocalPort();
	}

	
	public InetAddress getAddress() {
		return this.receiver.getLocalAddress();
	}

	
	public int getBufferSize() {
		return this.bufferSize;
	}
}
