/**
 * 
 */
package edu.dhbw.mannheim.tigers.sim.util.network;


import java.io.IOException;
import java.net.BindException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;

import org.apache.log4j.Logger;


/**
 * This {@link ITransmitter} implementation is capable of sending a {@code byte[]}-data somewhere via UDP.
 * 
 * @author Gero
 * 
 */
public class UnicastUDPTransmitter implements ITransmitter<byte[]> {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	protected final Logger	log			= Logger.getLogger(getClass());
	

	// Communication
	private DatagramSocket	socket		= null;
	private InetAddress		targetAddr;
	private final int			targetPort;
	
	/** The internal state-switch of this transmitter */
	private boolean			readyToSend	= false;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see UnicastUDPTransmitter
	 * @param localPort
	 * @param localAddrStr
	 * @param targetAddrStr
	 * @param targetPort
	 * @throws BindException
	 */
	public UnicastUDPTransmitter(int localPort, String localAddrStr, String targetAddrStr, int targetPort) throws BindException {
		this.targetPort = targetPort;

		InetAddress localAddr = null;
		try {
			localAddr = InetAddress.getByName(localAddrStr);
		} catch (UnknownHostException err) {
			log.error("The local address could not be resolved!", err);
		}
		
		while (socket == null) {
			try {
				this.socket = new DatagramSocket(localPort, localAddr);
			} catch (BindException err) {
				throw err;
			} catch (SocketException err) {
				log.info("Port " + localPort + " used, will try " + ++localPort + " instead!");
				continue;
			}
		}
		
		try {
			this.targetAddr = InetAddress.getByName(targetAddrStr);
		} catch (UnknownHostException err) {
			log.error("The Host could not be found!", err);
		}
		
		synchronized (this) {
			readyToSend = true;
		}
	}
	
	
	/**
	 * @see UnicastUDPTransmitter
	 * @param localPort
	 * @param targetAddr
	 * @param targetPort
	 * @throws BindException 
	 */
	public UnicastUDPTransmitter(int localPort, String targetAddr, int targetPort) throws BindException {
		this(localPort, "*", targetAddr, targetPort);
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public boolean send(byte[] data) {
		// Synchronize access to socket as it belongs to the 'state'
		DatagramSocket socket = null;
		synchronized (this) {
			if (!isReady()) {
				log.error("Transmitter is not ready to send!");
				return false;
			}
			
			socket = this.socket;
		}
		
		DatagramPacket packet = new DatagramPacket(data, data.length, targetAddr, targetPort);
		
		// Receive _outside_ the synchronized state, to prevent blocking of the state
		try {
			socket.send(packet); // DatagramPacket is sent...
		} catch (IOException err) {
			log.error("Error while sending data to: '" + targetAddr + ":" + targetPort + "'!");	//, err);
			return false;
		}
		
		return true;
	}
	

	@Override
	public synchronized void cleanup() {
		readyToSend = false;
		
		if (socket != null) {
			socket.close();
			socket = null;
		}
	}
	

	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public synchronized int getLocalPort() {
		return readyToSend ? socket.getLocalPort() : UNDEFINED_PORT;
	}
	

	@Override
	public synchronized InetAddress getLocalAddress() {
		return targetAddr;
	}
	

	@Override
	public synchronized int getTargetPort() {
		return targetPort;
	}
	

	@Override
	public synchronized boolean isReady() {
		return readyToSend;
	}
	

	synchronized DatagramSocket getSocket() {
		return socket;
	}
}
