package edu.dhbw.mannheim.tigers.sim.util.observer;


import java.util.LinkedList;


/**
 * <p>
 * Implements the observable-part of the Observer-Pattern.
 * </p>
 * 
 * @author Gero
 * 
 * @param <S> Observable type (Subject)
 * @param <O> Observer type
 * @param <E> Event type
 */
public class Observable<S extends IObservable<S, O, E>, O extends IObserver<S, O, E>, E> implements
		IObservable<S, O, E> {

	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final LinkedList<O>	registeredObservers;
	private final Object				listSync	= new Object();
	
	private E							lastEvent;
	

	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Adds the given observer to registeredObservers
	 * 
	 * @param observers: The {@link IObserver} that should listen to this observable.
	 */
	public Observable(E initEvent, O... observers) {
		registeredObservers = new LinkedList<O>();
		
		this.lastEvent = initEvent;
		
		addObservers(observers);
	}
	

	public Observable(E initEvent) {
		registeredObservers = new LinkedList<O>();
		
		this.lastEvent = initEvent;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Notifies all registered observers.
	 * 
	 * @param event: The {@link Event} that changed the model
	 */
	@Override
	public void notifyObservers(E event) {
		if (event == null) {
			return;
		}
		
		synchronized (listSync) {
			lastEvent = event;
			
			doNotify(event);
		}
	}
	

	@Override
	public void notifyObservers() {
		
		synchronized (listSync) {
			doNotify(lastEvent);
		}
	}
	

	private void doNotify(E event) {
		for (O observer : registeredObservers) {
			observer.update(this, event);
		}
	}
	


	// --------------------------------------------------------------------------
	// --- observers ------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public boolean addObservers(O... observers) {
		synchronized (listSync) {
			boolean result = true;
			
			for (O obs : observers) {
				if (obs == null || !registeredObservers.add(obs)) {
					result = false;
				}
			}
			
			notifyObservers(lastEvent);
			
			return result;
		}
	}
	

	@Override
	public boolean removeObserver(O observer) {
		synchronized (listSync) {
			boolean temp = registeredObservers.remove(observer);
			return temp;
		}
	}
}
